import copy
import math
import os
import pprint

import torch
import logging
from line_profiler import profile
from transformers.generation.logits_process import (
    LogitsProcessor,
    LOGITS_PROCESSOR_INPUTS_DOCSTRING,
)
from transformers.utils import add_start_docstrings

logger = logging.getLogger(__name__)


class GrammarConstrainedLogitsProcessor(LogitsProcessor):
    def __init__(self, grammar_constraint, valid_token_start_idx=None):
        self.last_size = None
        self.grammar_constraint = grammar_constraint
        self.batch_parsing_states = None
        self.valid_token_start_idx = None

    def mask_logits(self, logits, device):
        masked_logits = logits.clone()
        # resolve each stack to a tensor of True/False for each token
        # indicating acceptance
        # acceptance = self.grammar_acceptor.filter_vocab(self.stacks, device)
        acceptance = self.grammar_constraint.batch_filter_vocab(
            self.batch_parsing_states, device
        )
        # acceptance is a tensor of shape (batch_size, vocab_size)
        # get the indices of the accepted tokens
        # do the following operation only in debug mode
        if os.getenv("DEBUG_MODE") == "True":
            # convert acceptance to numpy array
            batch_size, vocab_size = acceptance.shape
            acceptance_np = acceptance.cpu().numpy()
            accepted_x, accepted_y = acceptance_np.nonzero()
            # dict of {batch_index: [accepted_token_indices]}
            # initialize the dict with empty list
            accepted_token_indices = {i: [] for i in range(batch_size)}
            for x, y in zip(accepted_x, accepted_y):
                accepted_token_indices[x].append(y)
            logger.debug("Accepted token indices for the current batch:")
            logger.debug("\n" + pprint.pformat(accepted_token_indices))
            # convert token_ids to tokens
            accepted_tokens = {
                i: [
                    self.grammar_constraint.tokenizer.decode([token_id])
                    for token_id in token_ids
                ]
                for i, token_ids in accepted_token_indices.items()
            }
            logger.debug("Accepted tokens for the current batch:")
            logger.debug("\n" + pprint.pformat(accepted_tokens))
        # Logits to -inf where False
        masked_logits[~acceptance] = -math.inf
        return masked_logits

    # TODO: batching
    def process_logits(self, input_ids, scores):
        """
        :param input_ids:
        :param scores:
        :return:
        """
        # we dynamically create stacks at the first call, so that we know the batch size and beam size
        if self.batch_parsing_states is None:
            self.batch_parsing_states = [
                # self.grammar_constraint.init_stacks()
                copy.deepcopy(
                    self.grammar_constraint.string_recognizer.get_initial_parsing_state()
                )
                for _ in range(len(input_ids))
            ]

        if os.getenv("DEBUG_MODE") == "True":
            print("-" * 80)

        logger.debug("input_ids: \n" + pprint.pformat(input_ids))
        # logger.debug("scores: \n" + pprint.pformat(scores))
        logger.debug("last_size: \n" + pprint.pformat(self.last_size))
        logger.debug(
            "num of stacks: \n"
            + pprint.pformat(
                [len(acc_state.stacks) for acc_state in self.batch_parsing_states]
            )
        )
        # logger.debug("stacks: \n" + pprint.pformat(self.batch_parsing_states.stacks))

        self.batch_parsing_states = (
            self.grammar_constraint.update_state_with_batch_token_seqs(
                input_ids, self.batch_parsing_states, self.valid_token_start_idx
            )
        )
        logger.debug(f"input_ids: {input_ids}")

        masked_scores = self.mask_logits(scores, scores.device)
        return masked_scores

    @add_start_docstrings(LOGITS_PROCESSOR_INPUTS_DOCSTRING)
    @profile
    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor
    ) -> torch.FloatTensor:
        return self.process_logits(input_ids, scores)



import unittest

from transformers import AutoTokenizer

from tests.test_accept_token_sequence._test_accept_tokens_mixin import (
    TokenizerTesterMixin,
)


class Gemma2TokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = AutoTokenizer
    pretrained_name = "google/gemma-2-2b-it"

    def setUp(self):
        super().setUp()



from unittest import TestCase

from transformers_cfg.parser import parse_ebnf
from transformers_cfg.recognizer import StringRecognizer
from dataclasses import dataclass


@dataclass
class MoleculeTestCase:
    name: str
    molecule: str


valid_smiles_sentences = [
    MoleculeTestCase("simple_atom", "C"),
    MoleculeTestCase("single_bond_no_hyphen", "CC"),
    MoleculeTestCase("double_bond", "C=O"),
    MoleculeTestCase("dot", "C.O"),
    MoleculeTestCase("radical", "CC(C)C"),
    MoleculeTestCase("isotope", "[14c]"),
    MoleculeTestCase("aromatic_no_hyphen", "C1CC1"),
    MoleculeTestCase("interleaved_cycle_explicit", "C1=CC=CC=C1"),
    MoleculeTestCase("interleaved_cycle_colon", "C1:C:C1"),
    MoleculeTestCase("interleaved_cycle_lower_case", "c1cc1"),
    MoleculeTestCase("cis_bond_right", "F/C=C\\F"),
    MoleculeTestCase("trans_bond_left", "F\\C=C\\F"),
    MoleculeTestCase("d_alanine", "C[C@H](C(=O)O)N"),
    MoleculeTestCase("l_alanine", "C[C@@H](C(=O)O)N"),
    MoleculeTestCase("nested_cycles", "C12(CCC1)CCC2"),
    MoleculeTestCase("charge", "[Cu+2].[O-]S(=O)[O-]"),
    MoleculeTestCase("mix_of_cases", "Cc(cc1)ccc1C#N"),
    MoleculeTestCase("mix_of_bonds_and_cycles", "C1CC/C=C1/C=C/C=C/C2=C(C)/CCC2"),
    MoleculeTestCase("wildcard", "Oc1c(*)cccc1"),
]

valid_smiles_prefixes = [
    MoleculeTestCase("empty_string", ""),
    MoleculeTestCase("simple_atom_dangling_bond", "C#"),
    MoleculeTestCase("unbalanced_paranthesis", "C(C(C)"),
    # Failure cases of SMILES in general
    # MoleculeTestCase("lowercase_outside_cycle", "c"),
    # MoleculeTestCase("unclosed cycle", "C1CC/C=C1/C=C/C=C/C2=C(C)/CCC"),
    # MoleculeTestCase("unclosed cycle", "C1CCC"),
]

invalid_smiles_sentences = [
    MoleculeTestCase("fake_atom", "L"),
    MoleculeTestCase("fake_molecule_in_brackets", "[Xx]"),
    MoleculeTestCase("bond_outside_parentheses", "CCC=(O)O"),
    MoleculeTestCase("double_double_bond", "C==C"),
    MoleculeTestCase("empty_paranthesis", "()"),
    MoleculeTestCase("invalid_charge", "[Cu+20].[O-]S(=O)(=O)[O-]"),
    # Failure cases of SMILES in general
    # MoleculeTestCase("two_bonds_same_atom", "C12C2CCC1"),
    # MoleculeTestCase("self-bond", "C11"),
]

valid_isocyanite_sentences = [
    MoleculeTestCase("short_isocyanate", "O=C=NCCCCCCN=C=O"),
    MoleculeTestCase("right_group", "CC1=C(C=C(C=C1)CN=C=O)N=C=O"),
    MoleculeTestCase("trans_bond_right", "Cc1ccc(cc1\\N=C=O)\\N=C=O"),
    MoleculeTestCase("group_radical", "O=C=NC1CCC(CC2CCC(CC2)N=C=O)CC1"),
    MoleculeTestCase("trans_bond_left", "O=C=N\\C1CC(C\\N=C=O)(CC(C1)(C)C)C"),
    MoleculeTestCase("trans_bond", "O=C=N\\CCCCCC/N=C=O"),
    MoleculeTestCase("group_radicals", "CCOC(C(N=C=O)CCCCN=C=O)=O"),
    MoleculeTestCase(
        "simple_atom", "O=C=NC1=CC=CC(CC2=CC=C(C=C2N=C=O)CC3=CC=C(C=C3)N=C=O)=C1"
    ),
    MoleculeTestCase(
        "single_bond_no_hyphen",
        "O=C=NC1=CC(CC2=C(C=C(C=C2)CC3=CC=C(C=C3N=C=O)CC4=CC=C(C=C4)N=C=O)N=C=O)=CC=C1",
    ),
    MoleculeTestCase(
        "double_bond",
        "O=C=NC1=CC=C(C=C1)CC2=CC=C(C=C2N=C=O)CC3=C(C=C(C=C3)CC4=CC=C(C=C4N=C=O)CC5=CC=C(C=C5)N=C=O)N=C=O",
    ),
    MoleculeTestCase("interleaved_cycle_explicit", "CC1(CC(CC(CN=C=O)(C1)C)N=C=O)C"),
    MoleculeTestCase("interleaved_cycle_colon", "CC1=C(C=C(C=C1)CN=C=O)N=C=O"),
    MoleculeTestCase("cycles", "O=C=N\\c1ccc(cc1)Cc2ccc(\\N=C=O)cc2"),
]

valid_acrylate_sentences = [
    MoleculeTestCase("simple_acrylate", "COC(=O)C=C"),
    MoleculeTestCase("simple_acrylate", "C=CC(=O)OC1=CC=CC=C1"),
    MoleculeTestCase("simple_acrylate_group_variation", "CC(=C)C(=O)OC1=CC=CC=C1"),
    MoleculeTestCase("", "C=CC(=O)OCCC1=CC=CC=C1"),
    MoleculeTestCase("", "CCC(C)OC(=O)C=C"),
    MoleculeTestCase("", "C=CC(=O)OC1=C(C(=C(C(=C1F)F)F)F)F"),
    MoleculeTestCase("", "CC(C)COC(=O)C(=C)C"),
    MoleculeTestCase("", "CCC(C)OC(=O)C(=C)C"),
    MoleculeTestCase("", "CCCOC(=O)C(=C)C"),
    MoleculeTestCase("", "CC1CC(CC(C1)(C)C)OC(=O)C(=C)C"),
    MoleculeTestCase("", "CCCOC(=O)C=C"),
    MoleculeTestCase("", "COCCOC(=O)C=C"),
    MoleculeTestCase("", "CC(=C)C(=O)OCCOC1=CC=CC=C1"),
    MoleculeTestCase("", "CCCCCCOC(=O)C=C"),
    MoleculeTestCase("", "CCCCOCCOC(=O)C(=C)C"),
    MoleculeTestCase("", "CC(=C)C(=O)OC"),
    MoleculeTestCase("", "CCCCOC(=O)C=C"),
    MoleculeTestCase("", "CCOCCOC(=O)C(=C)C"),
    MoleculeTestCase("", "CC(=C)C(=O)OC1CC2CCC1(C2(C)C)C"),
    MoleculeTestCase("", "CCCCC(CC)COC(=O)C(=C)C"),
    MoleculeTestCase("", "CC(C)(COCCCOC(=O)C=C)COCCCOC(=O)C=C"),
    MoleculeTestCase("", "C=CC(=O)OCCCCCCOC(=O)C=C"),
    MoleculeTestCase("", "C=CC(=O)OCC(CO)(COC(=O)C=C)COC(=O)C=C"),
    MoleculeTestCase("", "CCC(COCCCOC(=O)C=C)(COCCCOC(=O)C=C)COCCCOC(=O)C=C"),
    MoleculeTestCase("", "CCC(COCC(CC)(COC(=O)C=C)COC(=O)C=C)(COC(=O)C=C)COC(=O)C=C"),
    MoleculeTestCase(
        "", "C=CC(=O)OCC(CO)(COCC(COC(=O)C=C)(COC(=O)C=C)COC(=O)C=C)COC(=O)C=C"
    ),
    MoleculeTestCase(
        "", "C=CC(=O)OCC(COCC(COC(=O)C=C)(COC(=O)C=C)COC(=O)C=C)(COC(=O)C=C)COC(=O)C=C"
    ),
]

valid_chain_extender_sentences = [
    MoleculeTestCase("simplest_chain_extender", "OCCO"),
    MoleculeTestCase("", "OC(C)CCO"),
    MoleculeTestCase("", "OCCOCO"),
    MoleculeTestCase("", "OCCNC(=O)NCCCCCCNC(=O)NCCO"),
    MoleculeTestCase("", "OC(=O)C(N)CCCCN"),
    MoleculeTestCase("", "Oc1ccc(cc1)CCC(=O)OCCOC(=O)CCc1ccc(cc1)O"),
    MoleculeTestCase("", "OC(=O)C(N)CCN"),
    MoleculeTestCase("", "N1CCNCC1"),
    MoleculeTestCase("", "Nc1ccc(cc1)SSc2ccc(cc2)N"),
    MoleculeTestCase("", "Nc1ccc(cc1)Cc2ccc(cc2)N"),
]

TestCases = {
    "generic": (
        valid_smiles_sentences,
        valid_smiles_prefixes,
        invalid_smiles_sentences,
    ),
    "isocyanates": (
        valid_isocyanite_sentences,
        valid_smiles_sentences,
        invalid_smiles_sentences,
    ),
    "acrylates": (
        valid_acrylate_sentences,
        valid_smiles_sentences,
        invalid_smiles_sentences,
    ),
    "chain_extenders": (
        valid_chain_extender_sentences,
        valid_smiles_sentences,
        invalid_smiles_sentences,
    ),
}


class Test_parsing_smiles_object(TestCase):
    def setUp(self):
        self.recognizers = {}
        for grammar_name in TestCases:
            with open(f"examples/grammars/SMILES/{grammar_name}.ebnf", "r") as file:
                input_text = file.read()
            parsed_grammar = parse_ebnf(input_text)

            start_rule_id = parsed_grammar.symbol_table["root"]

            self.recognizers[grammar_name] = StringRecognizer(
                parsed_grammar.grammar_encoding, start_rule_id
            )
        print("SetUp successfull!", flush=True)

    def test_valid_sentence(self):
        for grammar_name, recognizer in self.recognizers.items():
            valid_full, valid_prefix, invalid = TestCases[grammar_name]

            for molecule_test_case in valid_full:
                fail_msg = (
                    f"{grammar_name.capitalize()}:"
                    + f"Failed on {molecule_test_case.name}, {molecule_test_case.molecule}"
                )
                print(fail_msg, flush=True)
                self.assertEqual(
                    True,
                    recognizer._accept_string(molecule_test_case.molecule),
                    msg=fail_msg,
                )
            for molecule_test_case in valid_prefix + invalid:
                fail_msg = (
                    f"{grammar_name.capitalize()}:"
                    + f"Failed on {molecule_test_case.name}, {molecule_test_case.molecule}"
                )
                print(fail_msg, flush=True)
                self.assertEqual(
                    False,
                    recognizer._accept_string(molecule_test_case.molecule),
                    msg=fail_msg,
                )

    def test_valid_prefixes(self):
        for grammar_name, recognizer in self.recognizers.items():
            valid_full, valid_prefix, invalid = TestCases[grammar_name]
            for molecule_test_case in valid_full + valid_prefix:
                fail_msg = (
                    f"{grammar_name.capitalize()}:"
                    + f"Failed on {molecule_test_case.name}, {molecule_test_case.molecule}"
                )
                print(fail_msg, flush=True)
                self.assertEqual(
                    True,
                    recognizer._accept_prefix(molecule_test_case.molecule),
                    msg=fail_msg,
                )

            for molecule_test_case in invalid:
                fail_msg = (
                    f"{grammar_name.capitalize()}:"
                    + f"Failed on {molecule_test_case.name}, {molecule_test_case.molecule}"
                )
                print(fail_msg, flush=True)
                self.assertEqual(
                    False,
                    recognizer._accept_prefix(molecule_test_case.molecule),
                    msg=fail_msg,
                )



from dataclasses import dataclass
from typing import Tuple

from dataclasses import dataclass


@dataclass
class PartialUTF8:
    """
    A data class representing the state of a partially decoded UTF-8 sequence.

    Attributes:
    - value (int): The current accumulated value of the partially decoded Unicode code point.
                   This attribute stores the bits that have been decoded so far. For a fully decoded
                   character or before any partial decoding has started, this would typically be `0`.

    - n_remain (int): The number of bytes remaining to complete the current UTF-8 encoded character.
                      A value of `-1` indicates that there is no ongoing partial decoding, i.e.,
                      either decoding has not started, or the last character was fully decoded.

    This class is used to handle situations where UTF-8 encoded data may end in the middle of a character
    sequence, allowing for the decoding process to be resumed when more data becomes available.
    """

    value: int = 0  # Default to 0, indicating no partial value accumulated
    n_remain: int = (
        -1
    )  # Default to -1, indicating no bytes are currently expected to complete the character

    def __hash__(self):
        return hash((self.value, self.n_remain))

    def __eq__(self, other):
        if not isinstance(other, PartialUTF8):
            return NotImplemented
        return self.value == other.value and self.n_remain == other.n_remain


from typing import List, Tuple
from functools import lru_cache


@lru_cache(maxsize=3000000)
def decode_utf8(
    src: bytes, partial_start: PartialUTF8
) -> Tuple[List[int], PartialUTF8]:
    # Lookup table for determining the total bytes based on the first byte's high 4 bits
    lookup = [1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 2, 2, 3, 4]
    pos = 0  # Position in the src bytes to start decoding from
    code_points = []  # List to store the decoded Unicode code points
    value = partial_start.value  # Start with any previously partial decoded value
    n_remain = partial_start.n_remain  # Number of bytes remaining from a partial decode

    # If there's a partial sequence left from last decode, try to continue decoding it
    while pos < len(src) and n_remain > 0:
        next_byte = src[pos]  # Get the next byte to process
        # Check if the continuation byte format is correct (`10xxxxxx`)
        if (next_byte >> 6) != 2:
            # If not, it's an invalid sequence. Abort and return a special error state.
            code_points = [0]
            return code_points, PartialUTF8(0, -1)

        # Accumulate the value by shifting left and adding the relevant 6 bits
        value = (value << 6) + (next_byte & 0x3F)
        pos += 1  # Move to the next byte
        n_remain -= 1  # Decrement the number of remaining bytes

    # If we've completed a partial sequence, add its value to the code points
    if partial_start.n_remain > 0 and n_remain == 0:
        code_points.append(value)

    # Process the rest of src as complete or new UTF-8 sequences
    while pos < len(src):
        first_byte = src[pos]  # Get the first byte of the next sequence
        highbits = first_byte >> 4  # Extract the high 4 bits for the lookup table
        n_remain = lookup[highbits] - 1  # Determine remaining bytes in this sequence

        # If lookup returns an invalid number, it's an invalid sequence. Abort.
        if n_remain < 0:
            # raise ValueError("Invalid UTF-8 sequence")
            code_points = [0]
            return code_points, PartialUTF8(0, -1)

        # Calculate the mask to isolate significant bits from the first byte
        mask = (1 << (7 - n_remain)) - 1
        value = first_byte & mask  # Apply the mask to get the initial value
        pos += 1  # Move to the next byte

        # Process the continuation bytes
        while pos < len(src) and n_remain > 0:
            next_byte = src[pos]
            # Shift the accumulated value and add the next 6 significant bits
            value = (value << 6) + (next_byte & 0x3F)
            pos += 1  # Move to the next byte
            n_remain -= 1  # Decrement the number of remaining bytes

        # If the sequence is complete, add its decoded value to the code points
        if n_remain == 0:
            code_points.append(value)

    # # Append a terminating value to indicate the end (following llama-cpp implementation)
    # code_points.append(0)
    # the following line is crucial for LRU cache to work, as it reset to the initial state
    if n_remain == 0:
        n_remain = -1
        value = 0

    # Return the decoded code points and the state of any partial decoding
    return code_points, PartialUTF8(value, n_remain)


def decode_utf8_leading_char(src: bytes) -> tuple:
    first_byte = src[0]
    highbits = first_byte >> 4
    lookup = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 4]
    char_len = lookup[highbits]

    # Extract the relevant bytes for the UTF-8 character
    utf8_char_bytes = src[:char_len]

    # Decode the character
    char = utf8_char_bytes.decode("utf-8")

    # Use ord() to convert the single character to its Unicode code point
    code_point = ord(char)

    # Remaining bytes
    remaining_bytes = src[char_len:]

    return code_point, remaining_bytes


def decode_utf8_string(utf8_bytes: bytes) -> list:
    code_points = []
    while utf8_bytes:
        code_point, utf8_bytes = decode_utf8_leading_char(utf8_bytes)
        code_points.append(code_point)
    return code_points


if __name__ == "__main__":
    # Given string
    my_string = "€Hello"  # The Euro symbol followed by "Hello"

    # Get UTF-8 encoded bytes
    utf8_bytes = my_string.encode("utf-8")

    assert utf8_bytes == b"\xe2\x82\xacHello"

    # Example usage with the Euro symbol followed by more characters
    code_point, remaining_bytes = decode_utf8_leading_char(utf8_bytes)

    print(f"Code Point: {code_point}")  # Expected Output: 8364 (Euro symbol)
    print(f"Remaining Bytes: {remaining_bytes}")  # Expected Output: b'Hello'

    # Example usage with the entire string
    code_points = decode_utf8_string(utf8_bytes)

    print(
        f"Code Points: {code_points}"
    )  # Expected Output: [8364, 72, 101, 108, 108, 111]

    print("-" * 50)

    # Example usage:
    utf8_bytes = b"\xe2\x82\xacHello"  # UTF-8 encoded string (Euro symbol + "Hello")
    partial_start = PartialUTF8()  # Assuming start with no partial sequence
    code_points, partial_utf8 = decode_utf8(utf8_bytes, partial_start)

    print("Code Points:", code_points)
    print("Remaining UTF-8 State:", partial_utf8.value, partial_utf8.n_remain)



import json
from typing import List

from termcolor import colored


def ints2bytes(sequence: List[int]) -> bytes:
    # check in the range of 0-255
    for item in sequence:
        if not 0 <= item <= 255:
            raise ValueError(f"item: {item} is not in the range [0, 255]")
    return bytes(sequence)


def bytes2ints(byte_sequence: bytes) -> List[int]:
    return list(byte_sequence)


def intervals_intersect(low1, high1, low2, high2):
    """
    Check if two intervals [low1, high1] and [low2, high2] intersect.

    :param high1: High bound of the first interval.
    :param low1: Low bound of the first interval.
    :param high2: High bound of the second interval.
    :param low2: Low bound of the second interval.
    :return: True if the intervals intersect, False otherwise.
    """
    # Check if one interval is completely to the right of the other
    if low1 > high2 or low2 > high1:
        return False

    # If the above condition is not met, the intervals intersect
    return True


def pprint_token_ids(tokenizer, token_ids=None, text=None):
    if token_ids is None and text is None:
        raise ValueError("Either token_ids or text should be provided")
    if token_ids is None:
        token_ids = tokenizer.encode(text, add_special_tokens=False)
    special_token_ids = tokenizer.all_special_ids
    special_tokens = tokenizer.all_special_tokens
    special_id2token = {
        id: token for id, token in zip(special_token_ids, special_tokens)
    }
    # loop over token_ids and color the special tokens
    colored_token_ids = []

    for token_id in token_ids:
        if token_id in special_id2token:
            colored_token_ids.append(colored(token_id, "red", attrs=["bold"]))
        else:
            colored_token_ids.append(str(token_id))
    colored_token_ids_str = [str(item) for item in colored_token_ids]
    print("[" + ", ".join(colored_token_ids_str) + "]")


def get_tokenizer_model_type(model: str = "gpt2"):
    """
    reference https://github.com/huggingface/transformers/blob/main/src/transformers/tokenization_utils_fast.py#L729
    :param model:
    :return: BPE, Unigram, WordPiece, WordLevel
    SentencePiece is used in conjunction with Unigram
    """
    from transformers import AutoTokenizer

    # if the tokenizer is not in the repo, it will raise OSError
    # OSError: Can't load tokenizer for 'xxx'
    # This happens when the model reuses the tokenizer of another model
    if type(model) == str:
        try:
            tokenizer = AutoTokenizer.from_pretrained(model, use_fast=True)
            # check if the tokenizer is fast
        except OSError:
            return None
    else:
        tokenizer = model

    if not tokenizer.is_fast:
        raise ValueError(f"The tokenizer {model} is not fast tokenizer")
    tokenizer_json = json.loads(tokenizer._tokenizer.to_str())
    model_type = tokenizer_json["model"]["type"]
    if (
        model_type == "BPE"
        and tokenizer_json["pre_tokenizer"] is not None
        and (
            tokenizer_json["pre_tokenizer"]["type"] == "ByteLevel"
            or (
                "pretokenizers" in tokenizer_json["pre_tokenizer"]
                and tokenizer_json["pre_tokenizer"]["pretokenizers"][1]["type"]
                == "ByteLevel"
            )
        )
    ):
        model_type = "ByteLevelBPE"
    return model_type



import logging
from abc import ABC
from functools import lru_cache
from typing import List, Optional

import torch

from transformers_cfg.recognizer import StringRecognizer, AcceptState
from transformers_cfg.parser import parse_ebnf
from transformers_cfg.tokenization.byte_trie import ByteTrie
from transformers_cfg.tokenization.middle.TokenizerMiddleMapping import (
    TokenizerMiddleMapping,
)

logger = logging.getLogger(__name__)


class AbsTokenRecognizer(ABC):
    def __init__(self, grammar_str, tokenizer, start_rule_name="root", unicode=False):
        parsed_grammar = parse_ebnf(grammar_str)
        grammar_encoding = parsed_grammar.grammar_encoding
        self.start_rule_id = parsed_grammar.symbol_table.get(start_rule_name)
        self.use_unicode = unicode

        self.eos_token_id = tokenizer.eos_token_id
        self.tokenizer = tokenizer
        self.string_recognizer = StringRecognizer(grammar_encoding, self.start_rule_id)
        self.byte_trie = ByteTrie.from_tokenizer(tokenizer)
        self.homomorphism = TokenizerMiddleMapping.from_hf_tokenizer(tokenizer)

    def try_accept_token_id(self, token_id: int, parsing_state: AcceptState) -> bool:
        stacks = parsing_state.stacks
        if self.string_recognizer._must_stop(stacks):
            if token_id == self.eos_token_id:
                return True
            else:
                return False
        if token_id == self.eos_token_id:
            if self.string_recognizer._can_stop(stacks):
                # if at least one of the stack is empty, we can stop
                # we clear all the stacks, meaning that we don't accept any token after EOS
                return True
            else:
                return False
        # for code_point in self.mapping.map(token_id):
        #     stacks = self.grammar._consume_char_code_point(code_point, stacks)
        bytes_or_codepoints = self.homomorphism.map(token_id, verbose=False)
        new_acc_state = self.string_recognizer._update_state_with_bytes(
            bytes_or_codepoints, parsing_state, verbose=False
        )
        return len(new_acc_state.stacks) > 0

    def update_state_with_batch_token_seqs(self, *args, **kwargs):
        """Process a list of tokens according to the grammar rules."""
        raise NotImplementedError

    def batch_filter_vocab(self, batch_parsing_states, device) -> torch.Tensor:
        batch_acceptance = []
        for parsing_state in batch_parsing_states:
            batch_acceptance.append(self.filter_vocab(parsing_state, device))
        return torch.stack(batch_acceptance)

    def filter_vocab(self, parsing_state, device) -> torch.Tensor:
        if not parsing_state.stacks:  # Check if stacks is empty
            # Handle the empty case: for example, return a tensor of False
            # The size of the tensor should match the size of your vocabulary
            vocab_size = len(self.homomorphism)
            logger.debug(f"Empty stack, sum of acceptance: {0}")
            # size of the vocab
            accepts = [False] * vocab_size
            accepts[self.eos_token_id] = True
            return torch.tensor(accepts, dtype=torch.bool, device=device)

        acceptance = self.get_next_token_acceptance(parsing_state, device)

        return acceptance

    def get_next_token_acceptance(self, parsing_state, device) -> torch.Tensor:
        raise NotImplementedError

    def validate_and_set_eos_acceptance(self, acceptance: torch.Tensor) -> torch.Tensor:
        if torch.any(acceptance) == 0:
            acceptance[self.eos_token_id] = True
        else:
            if acceptance[self.eos_token_id]:
                raise ValueError()
            acceptance[self.eos_token_id] = False
        return acceptance

    def accept_token_ids(self, token_ids, stacks) -> bool:
        """Accept a list of token IDs according to the grammar rules."""
        raise NotImplementedError


class IncrementalTokenRecognizer(AbsTokenRecognizer):
    def __init__(self, grammar_str, start_rule_name, tokenizer, unicode=False):
        super().__init__(grammar_str, tokenizer, start_rule_name, unicode)
        self.last_size = None

    def _update_state_with_token_id(
        self, token_id: int, parsing_state: AcceptState
    ) -> AcceptState:
        if self.string_recognizer._must_stop(parsing_state.stacks):
            if token_id == self.eos_token_id:
                return self.string_recognizer.get_termination_parsing_state()
            else:
                raise ValueError(
                    f"All stacks are empty, so the only token accepted is EOS({self.eos_token_id}), but got {token_id}.\
                        This error is likely due to the previous token not being accepted by the grammar."
                )
        if token_id == self.eos_token_id:
            if self.string_recognizer._can_stop(parsing_state.stacks):
                # if at least one of the stack is empty, we can stop
                # we clear all the stacks, meaning that we don't accept any token after EOS
                return self.string_recognizer.get_termination_parsing_state()
            else:
                raise ValueError(
                    f"At least one of the stack should be empty when EOS is reached. However, "
                    f"the stacks are {parsing_state.stacks}"
                )

        bytes_or_codepoints = self.homomorphism.map(token_id)
        parsing_state = self.string_recognizer._update_state_with_bytes(
            bytes_or_codepoints, parsing_state
        )
        return parsing_state

        # if self.last_size is not set (which would be the case when processing the first token).
        # In this case, do nothing.

    def update_state_with_batch_token_seqs(
        self, input_ids, batch_parsing_states, valid_token_start_idx=None
    ):

        if self.last_size is None:
            valid_prefix_tokens = [
                (
                    single_input_ids[valid_token_start_idx:]
                    if valid_token_start_idx is not None
                    else []
                )
                for single_input_ids in input_ids
            ]

            # self.grammar_acceptor.accept_token_ids(valid_prefix_tokens, self.stacks)
            batch_parsing_states = [
                self._update_state_with_single_token_seq(prefix, parsing_state)
                for prefix, parsing_state in zip(
                    valid_prefix_tokens, batch_parsing_states
                )
            ]
            #  if the length of the current input IDs (input_ids[0]) is exactly one more than self.last_size.
            #  This is expected in a scenario where inputs are processed incrementally, one token at a time.
        elif len(input_ids[0]) == self.last_size + 1:
            batch_parsing_states = [
                self._update_state_with_token_id(single_input_ids[-1], parsing_state)
                for single_input_ids, parsing_state in zip(
                    input_ids, batch_parsing_states
                )
            ]
            #  ensure that the input size is consistent with the expected incremental processing
            #  (i.e., one token at a time).
        else:
            # here we check if the input_ids are one token longer than the last time we processed
            # but we don't check if input_ids are actually valid.
            # Imagine a scenario where we generate 10 tokens, then we replace the 10 generated tokens with 10 new tokens.
            # In this case, the input_ids will be consistent with the last_size, but the input_ids are not valid.
            # However, should we really check if the input_ids are valid here?
            # If we do, then we need to reparse the whole input_ids at each call, which is not efficient.
            # Maybe we should just trust the user to provide valid input_ids?
            # The conclusion is that, we assume the input_ids are valid, and our generation will be correct.
            # If the input_ids are not valid, then the generation result will be wrong and we don't take responsibility for that.
            raise RuntimeError(
                "Input ID's length is inconsistent with the current state of "
                "the GrammarConstrainedLogitsProcessor. If you want to process "
                "another input sequence, please instantiate a new "
                "GrammarConstrainedLogitsProcessor."
            )
        self.last_size = len(input_ids[0])

        return batch_parsing_states

    def _update_state_with_single_token_seq(
        self,
        token_ids: List[int],
        parsing_state: Optional[AcceptState] = None,
        as_string=True,
    ):
        if parsing_state is None:
            parsing_state = self.string_recognizer.get_initial_parsing_state()
        if as_string:
            string = self.tokenizer.decode(token_ids)
            parsing_state = self.string_recognizer._update_state_with_string(
                string, parsing_state
            )
        else:
            for i, token_id in enumerate(token_ids):
                parsing_state = self._update_state_with_token_id(
                    token_id, parsing_state
                )
                if len(parsing_state.stacks) > 0:
                    cur_token_ids = token_ids[: i + 1]
                    logging.debug(f"{cur_token_ids} is accepted")
                    decoded_string = self.tokenizer.decode(cur_token_ids)
                    logging.debug(f"The decoded string is {decoded_string}")
        return parsing_state

    def accept_token_ids(self, token_ids, stacks=None, as_string=True) -> bool:
        output_state = self._update_state_with_single_token_seq(
            token_ids, stacks, as_string
        ).stacks
        return True if output_state else False

    def get_next_token_acceptance(self, parsing_state, device) -> torch.Tensor:
        acceptance_matrix = torch.cat(
            [
                self.get_next_token_acceptance_for_single_stack(
                    tuple(stack), parsing_state.partial_utf8, device
                )
                for stack in parsing_state.stacks
            ]
        )
        # Merge stacks: any True => True
        acceptance = acceptance_matrix.reshape(len(parsing_state.stacks), -1).any(dim=0)
        return acceptance

    @lru_cache(maxsize=32768)
    def get_next_token_acceptance_for_single_stack(self, stack, partial_utf8, device):
        # stack = list(stack)  # needs to come in as a tuple for lru_cache
        assert isinstance(stack, tuple)

        if self.use_unicode:

            accept_f = lambda x: self.string_recognizer._try_accept_bytes(
                x, {stack}, partial_utf8=partial_utf8
            )
            token_acceptance = self.byte_trie.get_next_token_acceptance(
                accept=accept_f, accept_eos=False, eos_token_id=self.eos_token_id
            )
        else:
            accepts = [False] * len(self.homomorphism)
            token_acceptance = check_token_acceptance_in_trie(
                self.byte_trie.root,
                [stack],
                self.string_recognizer,
                self.eos_token_id,
                accepts,
            )
        x = torch.tensor(token_acceptance, dtype=torch.bool, device=device)
        x_eos = self.validate_and_set_eos_acceptance(x)
        return x_eos


# def check_token_acceptance_in_trie(trie, stacks, grammar, eos_token_id, accepts):

#     for byte, next_trie in trie.items():
#         if byte == LEAF:
#             token_id = next_trie
#             if token_id != eos_token_id:
#                 # if the stacks is not empty, it means we can still continue to parse
#                 # so we should accept the token
#                 accepts[token_id] = bool(stacks)
#             continue

#         new_stacks = set()
#         for stk in stacks:
#             if not stk:
#                 continue

#             next_element_offset = stk[-1]
#             num_chars = grammar.grammar_encoding[next_element_offset]

#             if not grammar.char_acceptance_at_element(next_element_offset).get(
#                 byte, False
#             ):
#                 # if the current byte is not accepted by the current rule, we need to try next rule
#                 continue

#             next_element_offset += num_chars + 1
#             new_stack = list(stk[:-1])
#             if grammar.grammar_encoding[next_element_offset]:
#                 new_stack.append(next_element_offset)
#             new_stacks.update(grammar.expand_stack_head(tuple(new_stack)))

#         if new_stacks:
#             check_token_acceptance_in_trie(
#                 next_trie, new_stacks, grammar, eos_token_id, accepts
#             )

#     return accepts


def check_token_acceptance_in_trie(trie_node, stacks, grammar, eos_token_id, accepts):
    if trie_node.is_end_of_word:
        token_id = trie_node.token_id
        if token_id != eos_token_id:
            # if the stacks is not empty, it means we can still continue to parse
            # so we should accept the token
            accepts[token_id] = bool(stacks)

    for byte, next_trie_node in trie_node.children.items():
        new_stacks = set()
        for stk in stacks:
            if not stk:
                continue

            next_element_offset = stk[-1]
            num_chars = grammar.grammar_encoding[next_element_offset]

            if not grammar.char_acceptance_at_element(next_element_offset).get(
                byte, False
            ):
                # if the current byte is not accepted by the current rule, we need to try next rule
                continue

            next_element_offset += num_chars + 1
            new_stack = list(stk[:-1])
            if grammar.grammar_encoding[next_element_offset]:
                new_stack.append(next_element_offset)
            new_stacks.update(grammar.expand_stack_head(tuple(new_stack)))

        if new_stacks:
            check_token_acceptance_in_trie(
                next_trie_node, new_stacks, grammar, eos_token_id, accepts
            )

    return accepts


class NonIncrementalTokenSeqRecognizer(IncrementalTokenRecognizer):
    def __init__(self, grammar_str, start_rule_name, tokenizer, unicode=False):
        super().__init__(grammar_str, start_rule_name, tokenizer, unicode)

    def update_state_with_batch_token_seqs(
        self, input_ids, batch_parsing_states, valid_token_start_idx=None
    ):

        if self.last_size is None:
            valid_prefix_tokens = [
                (
                    single_input_ids[valid_token_start_idx:]
                    if valid_token_start_idx is not None
                    else []
                )
                for single_input_ids in input_ids
            ]

            # self.grammar_acceptor.accept_token_ids(valid_prefix_tokens, self.stacks)
            resulting_batch_parsing_states = [
                self._update_state_with_single_token_seq(token_ids, parsing_state)
                for token_ids, parsing_state in zip(
                    valid_prefix_tokens, batch_parsing_states
                )
            ]
            #  if the length of the current input IDs (input_ids[0]) is exactly one more than self.last_size.
            #  This is expected in a scenario where inputs are processed incrementally, one token at a time.
            self.last_size = len(input_ids[0])
        else:

            # loop over the input_ids after the last_size
            resulting_batch_parsing_states = []

            for single_input_ids, _ in zip(input_ids, batch_parsing_states):
                valid_input_ids = single_input_ids[self.last_size :]
                # for i, token_id in enumerate(valid_input_ids):
                # parsing_state = self._consume_token_id(token_id, parsing_state)
                parsing_state = self._update_state_with_single_token_seq(
                    valid_input_ids, parsing_state=None
                )
                if len(parsing_state.stacks) == 0:
                    raise ValueError("The input is not accepted")
                resulting_batch_parsing_states.append(parsing_state)

        return resulting_batch_parsing_states


if __name__ == "__main__":
    from transformers import AutoTokenizer

    with open("examples/grammars/japanese.ebnf", "r") as file:
        input_text = file.read()
    parsed_grammar = parse_ebnf(input_text)
    parsed_grammar.print()

    tokenizer = AutoTokenizer.from_pretrained("gpt2")

    tokenRecognizer = IncrementalTokenRecognizer(
        grammar_str=input_text,
        start_rule_name="root",
        tokenizer=tokenizer,
        unicode=True,
    )

    japanese = "トリーム"  # "こんにちは"
    token_ids = tokenizer.encode(japanese)
    # 13298, 12675, 12045, 254
    init_state = None
    state = tokenRecognizer._update_state_with_single_token_seq(
        token_ids, init_state, as_string=False
    )

    if state.stacks:
        print("The Japanese input is accepted")
    else:
        print("The Japanese input is not accepted")

    korean = "안녕하세요"
    token_ids = tokenizer.encode(korean)
    init_state = tokenRecognizer.string_recognizer.get_initial_parsing_state()

    try:
        state = tokenRecognizer._update_state_with_single_token_seq(
            token_ids,
            init_state,
            as_string=False,
        )
        if state.stacks:
            print("The Korean input is accepted")
        else:
            print("The Korean input is not accepted")
    except ValueError as e:
        print("The Korean input is not accepted")



import json


def is_json_parsable(string):
    try:
        json.loads(string)
        return True
    except json.JSONDecodeError:
        return False
    except Exception as e:
        # You might want to handle or log other exceptions as well
        return False



import warnings

from transformers import PreTrainedTokenizer
from transformers_cfg.token_grammar_recognizer import IncrementalTokenRecognizer


from transformers_cfg.utils import pprint_token_ids


class TokenizerTesterMixin:
    tokenizer_class = None
    pretrained_name = None
    rust_tokenizer_class = None
    test_slow_tokenizer = True
    test_rust_tokenizer = True
    space_between_special_tokens = False
    from_pretrained_kwargs = None
    from_pretrained_filter = None
    from_pretrained_vocab_key = "vocab_file"
    test_seq2seq = True

    # set to True to test a sentencepiece tokenizer
    test_sentencepiece = False

    # set to True to ignore casing when testing a sentencepiece tokenizer
    # test_sentencepiece must also be set to True
    test_sentencepiece_ignore_case = False

    def setUp(self):
        self.tokenizer = self.get_tokenizer()

    def get_tokenizer(self, **kwargs) -> PreTrainedTokenizer:
        use_fast = kwargs.pop("use_fast", True)
        return self.tokenizer_class.from_pretrained(
            self.pretrained_name, use_fast=use_fast, **kwargs
        )

    def test_json_parsable(self):
        # Test that we can load a JSON object
        with open("examples/grammars/json.ebnf", "r") as file:
            input_text = file.read()
        JsontokenRecognizer = IncrementalTokenRecognizer(
            grammar_str=input_text, start_rule_name="root", tokenizer=self.tokenizer
        )

        valid_json = '{"foo": "bar", "baz": "bat"}'
        token_ids = self.tokenizer.encode(valid_json)
        pprint_token_ids(self.tokenizer, token_ids)

        # check if there is unk token
        for token_id in token_ids:
            if token_id == self.tokenizer.unk_token_id:
                warnings.warn(
                    f"unk token found in input_token_ids: {token_ids}, skipping test"
                )
                return

        acc_state = JsontokenRecognizer._update_state_with_single_token_seq(
            token_ids, as_string=False
        )
        # the json object is complete, so the stacks should be empty
        self.assertTrue(
            acc_state.stacks == set() or acc_state.stacks == set(tuple()),
            f"stacks: {acc_state.stacks}, not empty",
        )

    def test_balanced_parentheses(self):
        # Test that we can recognize a balanced parentheses
        with open("examples/grammars/balanced_parentheses.ebnf", "r") as file:
            input_text = file.read()
        recognizer = IncrementalTokenRecognizer(
            grammar_str=input_text, start_rule_name="root", tokenizer=self.tokenizer
        )

        balanced_parentheses = "((((((((()))))))))"
        token_ids = self.tokenizer.encode(balanced_parentheses)
        pprint_token_ids(self.tokenizer, token_ids)

        # check if there is unk token
        for token_id in token_ids:
            if token_id == self.tokenizer.unk_token_id:
                warnings.warn(
                    f"unk token found in input_token_ids: {token_ids}, skipping test"
                )
                return
        parsing_state = recognizer._update_state_with_single_token_seq(
            token_ids, as_string=False
        )
        # the json object is complete, so the stacks should be empty
        self.assertTrue(
            parsing_state.stacks == set() or parsing_state.stacks == set(tuple()),
            f"stacks: {parsing_state.stacks}, not empty",
        )

    def test_forcing_sequence(self):

        string_to_force = "12345 678 90"

        grammar_str = f"""
        root ::= "{string_to_force}"

        """

        tokenRecognizer = IncrementalTokenRecognizer(
            grammar_str=grammar_str, start_rule_name="root", tokenizer=self.tokenizer
        )

        token_ids = self.tokenizer.encode(string_to_force)
        pprint_token_ids(self.tokenizer, token_ids)

        # check if there is unk token
        for token_id in token_ids:
            if token_id == self.tokenizer.unk_token_id:
                warnings.warn(
                    f"unk token found in input_token_ids: {token_ids}, skipping test"
                )
                return

        acc_state = tokenRecognizer._update_state_with_single_token_seq(
            token_ids, as_string=False
        )
        # the json object is complete, so the stacks should be empty
        self.assertTrue(
            acc_state.stacks == set() or acc_state.stacks == set(tuple()),
            f"stacks: {acc_state.stacks}, not empty",
        )

    def test_emoji(self):
        """
        Test that we can accept emoji
        """

        with open("examples/grammars/emoji.ebnf", "r") as file:
            input_text = file.read()

        tokenRecognizer = IncrementalTokenRecognizer(
            grammar_str=input_text, start_rule_name="root", tokenizer=self.tokenizer
        )

        emoji = "😀😄😂"
        token_ids = self.tokenizer.encode(emoji)
        pprint_token_ids(self.tokenizer, token_ids)

        # check if there is unk token
        for token_id in token_ids:
            if token_id == self.tokenizer.unk_token_id:
                warnings.warn(
                    f"unk token found in input_token_ids: {token_ids}, skipping test"
                )
                return

        accpetance = tokenRecognizer.accept_token_ids(token_ids, as_string=False)

        self.assertTrue(accpetance, f"emoji: {emoji} not accepted, but it should be")



import unittest

from transformers import BloomTokenizerFast

from tests.test_accept_token_sequence._test_accept_tokens_mixin import (
    TokenizerTesterMixin,
)


@unittest.skip("Bloom is not supported and will be removed")
class BloomTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = BloomTokenizerFast
    pretrained_name = "bigscience/bloom-560m"

    def setUp(self):
        super().setUp()



import unittest

from transformers import AutoTokenizer

from tests.test_accept_token_sequence._test_accept_tokens_mixin import (
    TokenizerTesterMixin,
)


@unittest.skip("Falcom is not supported and will be removed")
class FalconTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = AutoTokenizer
    pretrained_name = "tiiuae/falcon-7b"

    def setUp(self):
        super().setUp()



from transformers_cfg.tokenization.SUPPORTED_TOKENIZERS import SUPPORTED_TOKENIZERS
from .ByteProxyMapping import ByteProxyMapping, LLAMAByteProxyMapper
import logging
from transformers import (
    GPT2TokenizerFast,
    BartTokenizerFast,
    T5TokenizerFast,
    CodeGenTokenizerFast,
    LlamaTokenizerFast,
    PreTrainedTokenizerFast,
    GemmaTokenizerFast
)

from transformers_cfg.tokenization.utils import get_tokenizer_charset

log = logging.getLogger(__name__)


class TokenizerMiddleMapping:
    def __init__(self, tokenizer):
        self.eos_token_id = tokenizer.eos_token_id
        self.bos_token_id = tokenizer.bos_token_id
        self.tokenizer = tokenizer
        self.special = tokenizer.all_special_ids
        self._length = len(self.tokenizer.get_vocab())

    def __len__(self):
        return self._length

    def map(self, token_id: int, verbose=False) -> bytes:
        raise NotImplementedError("This method should be implemented in the subclass")

    @classmethod
    def from_hf_tokenizer(cls, hf_tokenizer):
        assert (
            type(hf_tokenizer) in SUPPORTED_TOKENIZERS
        ), f"Tokenizer not supported: {hf_tokenizer.__class__.__name__}, supported tokenizers: {SUPPORTED_TOKENIZERS}"
        if isinstance(
            hf_tokenizer, (GPT2TokenizerFast, BartTokenizerFast, CodeGenTokenizerFast)
        ):
            return GPT2TokenizerMiddleMapping(hf_tokenizer)
        elif isinstance(hf_tokenizer, (LlamaTokenizerFast, GemmaTokenizerFast)):
            # deepseek, though inheriting from LlamaTokenizerFast, is actually a GPT2TokenizerFast
            # check https://github.com/epfl-dlab/transformers-CFG/issues/72
            if hf_tokenizer.name_or_path.startswith("deepseek-ai/deepseek-coder"):
                return GPT2TokenizerMiddleMapping(hf_tokenizer)
            return LLAMA1TokenizerMiddleMapping(hf_tokenizer)
        elif isinstance(hf_tokenizer, T5TokenizerFast):
            return T5TokenizerMiddleMapping(hf_tokenizer)
        elif isinstance(
            hf_tokenizer, PreTrainedTokenizerFast
        ) and hf_tokenizer.name_or_path.startswith("meta-llama/Meta-Llama-3"):
            return GPT2TokenizerMiddleMapping(hf_tokenizer)

    @staticmethod
    def auto_infer(hf_tokenizer):
        "beta version, not sure if it will work for all cases"
        charset = get_tokenizer_charset(hf_tokenizer)
        size = len(charset)
        if size >= 256 and size < 256 + 30:
            return GPT2TokenizerMiddleMapping(hf_tokenizer)
        elif "▁" in charset:
            return LLAMA1TokenizerMiddleMapping(hf_tokenizer)
        else:
            raise NotImplementedError(
                f"Tokenizer not supported: {hf_tokenizer.__class__.__name__}"
            )


class GPT2TokenizerMiddleMapping(TokenizerMiddleMapping):
    def __init__(self, tokenizer):
        super().__init__(tokenizer)
        self.byte_proxy_mapper = ByteProxyMapping(tokenizer)

    def map2proxy_token(self, token_id: int) -> str:
        # This is the case for BOS,
        if token_id in self.special:
            return ""
        # if token_id is tensor, convert it to int
        if hasattr(token_id, "item"):
            token_id = token_id.item()
        proxy_token = self.tokenizer.convert_ids_to_tokens(token_id)
        return proxy_token

    def map(self, token_id: int, verbose=False) -> bytes:
        proxy_token = self.map2proxy_token(token_id)
        if verbose:
            log.debug(f"token_id: {token_id}, token: {proxy_token}")

        return self.byte_proxy_mapper.map(proxy_token)


class LLAMA1TokenizerMiddleMapping(TokenizerMiddleMapping):
    def __init__(self, tokenizer):
        super().__init__(tokenizer)
        self.last_token_id = None
        self.byte_proxy_mapper = LLAMAByteProxyMapper()

    def map(self, token_id: int, verbose=False) -> bytes:
        # we need to check if the token is at the beginning of the sentence to remove the space
        # specific to BPE
        at_bos = False
        if self.last_token_id is not None and self.last_token_id == self.bos_token_id:
            at_bos = True
        self.last_token_id = token_id

        # This is the case for BOS,
        if token_id in self.special:
            return b""
        # if token_id is tensor, convert it to int
        if hasattr(token_id, "item"):
            token_id = token_id.item()
        proxy_token = self.tokenizer.convert_ids_to_tokens(token_id)

        token_bytes = self.byte_proxy_mapper.map(proxy_token)

        # check if the first byte is a space
        if token_bytes[0] == 32 and at_bos:
            # remove space at the beginning of the sentence
            token_bytes = token_bytes[1:]

        return token_bytes


class T5TokenizerMiddleMapping(TokenizerMiddleMapping):
    def __init__(self, tokenizer):
        super().__init__(tokenizer)
        self.at_bos = True
        self.byte_proxy_mapper = LLAMAByteProxyMapper()

    def map(self, token_id: int, verbose=False) -> bytes:
        # we need to check if the token is at the beginning of the sentence to remove the space
        # specific to BPE

        # This is the case for BOS,
        if token_id in self.special:
            self.at_bos = False
            return b""
        # if token_id is tensor, convert it to int
        if hasattr(token_id, "item"):
            token_id = token_id.item()
        proxy_token = self.tokenizer.convert_ids_to_tokens(token_id)

        token_bytes = self.byte_proxy_mapper.map(proxy_token)

        # check if the first byte is a space
        if token_bytes[0] == 32 and self.at_bos:
            # remove space at the beginning of the sentence
            token_bytes = token_bytes[1:]

        self.at_bos = False
        return token_bytes



import unittest
from transformers_cfg.utf8_utils import (
    decode_utf8,
    PartialUTF8,
)  # Make sure to import your function and class


class TestDecodeUTF8(unittest.TestCase):
    def test_decode_single_character(self):
        """Test decoding a single character."""
        utf8_bytes = b"\xe2\x82\xac"  # Euro sign
        expected_code_points = [
            8364,
        ]  # Euro sign code point followed by terminating 0
        result, _ = decode_utf8(utf8_bytes, PartialUTF8())
        self.assertEqual(result, expected_code_points)

    def test_decode_multiple_characters(self):
        """Test decoding a string with multiple UTF-8 characters."""
        utf8_bytes = b"Hello, \xe2\x82\xac!"  # "Hello, €!"
        expected_code_points = [
            72,
            101,
            108,
            108,
            111,
            44,
            32,
            8364,
            33,
        ]
        result, _ = decode_utf8(utf8_bytes, PartialUTF8())
        self.assertEqual(result, expected_code_points)

    def test_handle_incomplete_sequence(self):
        """Test handling of an incomplete UTF-8 sequence."""
        utf8_bytes = b"\xe2"  # Incomplete sequence for the Euro sign
        expected_code_points = []  # Expect a 0 due to incomplete sequence
        result, partial = decode_utf8(utf8_bytes, PartialUTF8())
        self.assertEqual(result, expected_code_points)
        starting_value = int.from_bytes(b"\xe2", "big")  # 226
        offset_value = int.from_bytes(b"\xe0", "big")  # 224
        self.assertEqual(partial.value, starting_value - offset_value)  # 226-224=2
        self.assertEqual(
            partial.n_remain, 2
        )  # Expect n_remain to be 2 because 2 more bytes are needed

    def test_continue_incomplete_sequence(self):
        """Test continuation of decoding with a previously incomplete sequence."""
        utf8_bytes = b"\x82\xac"  # Continuation of the Euro sign
        partial_start = PartialUTF8(
            value=2, n_remain=2
        )  # Simulate a previous state expecting 2 more bytes
        expected_code_points = [
            8364,
        ]  # Completed Euro sign code point
        result, _ = decode_utf8(utf8_bytes, partial_start)
        self.assertEqual(result, expected_code_points)

    def test_empty_string(self):
        """Test handling of an empty string."""
        utf8_bytes = b""
        expected_code_points = []
        result, _ = decode_utf8(utf8_bytes, PartialUTF8())
        self.assertEqual(result, expected_code_points)


if __name__ == "__main__":
    unittest.main()



from unittest import TestCase

from transformers_cfg.parser import parse_ebnf
from transformers_cfg.recognizer import StringRecognizer
from dataclasses import dataclass


@dataclass
class GeoQueryTestCase:
    name: str
    geo_query: str


valid_geo_query_sentences = [
    GeoQueryTestCase("simple_answer", "answer(smallest(state(all)))"),
    GeoQueryTestCase("state_id", "answer(highest(place(loc_2(stateid('hawaii')))))"),
    GeoQueryTestCase("river", "answer(river(all))"),
    GeoQueryTestCase("state", "answer(loc_1(major(river(all))))"),
    GeoQueryTestCase("next_to_2", "answer(state(next_to_2(stateid('texas'))))"),
    GeoQueryTestCase(
        "intersection",
        "answer(intersection(state(next_to_2(stateid('texas'))), loc_1(major(river(all)))))",
    ),
    GeoQueryTestCase("space in name", "answer(population_1(stateid('new york')))"),
    GeoQueryTestCase(
        "exclude",
        "answer(count(exclude(river(all), traverse_2(state(loc_1(capital(cityid('albany', _))))))))",
    ),
    GeoQueryTestCase(
        "city_id_with_state", "answer(population_1(cityid('washington', 'dc')))"
    ),
]

valid_geo_query_prefixes = [
    GeoQueryTestCase("empty_string", ""),
    GeoQueryTestCase("unbalanced_paranthesis", "answer(count(major(city(all"),
]

invalid_geo_query_sentences = [
    GeoQueryTestCase("no_answer", "highest(place(loc_2(stateid('kansas'))))"),
    GeoQueryTestCase("fake_country", "answer(major(city(loc_2(countryid('xx')))))"),
    GeoQueryTestCase("unexisting_function", "answer(population_2(stateid('hawaii')))"),
    GeoQueryTestCase("empty_operator", "answer(highest(place(loc_2())))"),
    GeoQueryTestCase("empty_paranthesis", "()"),
    GeoQueryTestCase(
        "missing_argument", "answer(intersection(state(next_to_2(stateid('texas'))), )"
    ),
]


class Test_parsing_geo_query_object(TestCase):
    def setUp(self):
        with open(f"examples/grammars/geo_query.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)
        start_rule_id = parsed_grammar.symbol_table["root"]
        self.recognizer = StringRecognizer(
            parsed_grammar.grammar_encoding, start_rule_id
        )
        print("SetUp successfull!", flush=True)

    def test_valid_sentence(self):

        for geo_query_test_case in valid_geo_query_sentences:
            self.assertEqual(
                True,
                self.recognizer._accept_string(geo_query_test_case.geo_query),
                msg=f"Failed on {geo_query_test_case.name}, {geo_query_test_case.geo_query}",
            )
        for geo_query_test_case in (
            valid_geo_query_prefixes + invalid_geo_query_sentences
        ):
            self.assertEqual(
                False,
                self.recognizer._accept_string(geo_query_test_case.geo_query),
                msg=f"Failed on {geo_query_test_case.name}, {geo_query_test_case.geo_query}",
            )

    def test_valid_prefixes(self):
        for geo_query_test_case in valid_geo_query_sentences + valid_geo_query_prefixes:
            self.assertEqual(
                True,
                self.recognizer._accept_prefix(geo_query_test_case.geo_query),
                msg=f"Failed on {geo_query_test_case.name}, {geo_query_test_case.geo_query}",
            )

        for geo_query_test_case in invalid_geo_query_sentences:
            self.assertEqual(
                False,
                self.recognizer._accept_prefix(geo_query_test_case.geo_query),
                msg=f"Failed on {geo_query_test_case.name}, {geo_query_test_case.geo_query}",
            )



import logging
from functools import lru_cache
from typing import List, Tuple, Set, Optional

from transformers_cfg.parser import (
    END_OF_RULE_MARKER,
    END_OF_ALTERNATE_MARKER,
    parse_ebnf,
    REF_RULE_MARKER,
)
from transformers_cfg.utf8_utils import PartialUTF8, decode_utf8
from transformers_cfg.utils import intervals_intersect
import logging


class AcceptState:
    def __init__(self, stacks, partial_utf8):
        self.stacks = stacks
        self.partial_utf8 = partial_utf8

    @staticmethod
    def empty_state():
        return AcceptState(set(), PartialUTF8())


class StringRecognizer:
    def __init__(
        self,
        grammar_encoding: List[int],
        start_rule_id: int = -1,
        rule_offsets: Optional[List[int]] = None,
        stacks: Optional[Set[Tuple[int]]] = None,
    ):
        # strictly speaking, we don't need to copy grammar_encoding because we don't modify it
        # but we do it anyway to be safe
        # in case where the grammar is very large, we can consider not copying it
        self.grammar_encoding = grammar_encoding
        if rule_offsets is not None:
            self.rule_offsets = rule_offsets
        else:
            if start_rule_id == -1:
                raise ValueError("start_rule_id cannot be None if rule_offsets is None")
            self.rule_offsets = self.init_rules(start_rule_id)
        # each stack is a list of indices into grammar_encoding
        # each index points to a rule's
        if stacks is not None:
            self.stacks = stacks
        else:
            if start_rule_id == -1:
                raise ValueError("start_rule_id cannot be None if stacks is None")
            self.stacks = self.init_stack(start_rule_id)
        self.start_rule_id = start_rule_id

    def init_rules(self, start_rule_id: int) -> List[int]:
        _rule_offset = 0
        rule_offsets: List = []
        # Build `rules` as an array of rule IDs to their positions in `grammar_src`
        while self.grammar_encoding[_rule_offset] != 0xFFFF:
            rule_id = self.grammar_encoding[_rule_offset]
            # store the offset idx
            if len(rule_offsets) <= rule_id:
                rule_offsets.extend([-1] * (rule_id - len(rule_offsets) + 1))
            rule_offsets[rule_id] = _rule_offset

            # Skip rule ID
            # _rule_offset += 1
            simple_rhs_offset = _rule_offset + 1

            # Skip rule alternates
            while self.grammar_encoding[simple_rhs_offset] != END_OF_RULE_MARKER:
                simple_rhs_offset = (
                    simple_rhs_offset + 1 + self.grammar_encoding[simple_rhs_offset]
                )

            # Skip 0 denoting end of rule
            # _rule_offset += 1
            _rule_offset = simple_rhs_offset + 1

        retrieved_start_rule_id = self.grammar_encoding[rule_offsets[start_rule_id]]
        assert retrieved_start_rule_id == start_rule_id

        return rule_offsets

    def init_stack(self, start_rule_id: int) -> Set[Tuple[int]]:

        stacks = set()
        # Loop over alternates of start rule to build initial stacks
        sub_rhs_offset = self.rule_offsets[start_rule_id] + 1
        while self.grammar_encoding[sub_rhs_offset]:
            stack: List[int] = []
            # If alternate is nonempty, add to stack
            element_offset = sub_rhs_offset + 1
            if self.grammar_encoding[element_offset] != END_OF_ALTERNATE_MARKER:
                stack.append(element_offset)
            stacks.update(self.expand_stack_head(tuple(stack)))
            sub_rhs_offset += 1 + self.grammar_encoding[sub_rhs_offset]
        return stacks

    def get_initial_parsing_state(self) -> AcceptState:
        return AcceptState(self.init_stack(self.start_rule_id), PartialUTF8())

    def get_termination_parsing_state(self) -> AcceptState:
        return AcceptState(set(), PartialUTF8())

    @lru_cache(maxsize=32768)
    def expand_stack_head(self, stack: Tuple[int]) -> Set[Tuple[int]]:
        """
        Stack is the internal state of the recognizer(Pushdown Automaton).
        This method updates the stack by advancing it to the next element.
        If the element is a non-terminal, it expands the stack by adding the elements of the referenced rule.
        A new stack is created for each alternate of the referenced rule, so we could have multiple stacks as output.
        :param stack:
        :return:
        """
        if len(stack) == 0:
            return {stack}

        # we get the last element of the stack, which is the element we are currently processing
        cur_element_offset = stack[-1]

        # if the element is a terminal, we don't need to advance the stack
        if self.grammar_encoding[cur_element_offset] != REF_RULE_MARKER:
            return {stack}
        # the remaining case is that the element is a non-terminal, i.e. a reference to another rule
        else:
            ref_rule_id = self.grammar_encoding[cur_element_offset + 1]
            # find the offset of the referenced rule
            ref_subrule_offset = self.rule_offsets[ref_rule_id] + 1
            new_stacks: Set[Tuple[int]] = set()
            # Loop over alternates of referenced rule to build new stacks
            while self.grammar_encoding[ref_subrule_offset] != END_OF_RULE_MARKER:
                # copy the original stack without the last element
                new_stack = list(stack[:-1])
                # if the rule ref is followed by another element, we add it to the stack
                next_element_offset = cur_element_offset + 2
                if (
                    self.grammar_encoding[next_element_offset]
                    != END_OF_ALTERNATE_MARKER
                ):
                    new_stack.append(next_element_offset)

                # if the referenced rule is not empty, we add its element offset to the stack
                ref_element_offset = ref_subrule_offset + 1
                if self.grammar_encoding[ref_element_offset] != END_OF_ALTERNATE_MARKER:
                    new_stack.append(ref_element_offset)

                new_stacks.update(self.expand_stack_head(tuple(new_stack)))
                ref_subrule_offset += self.grammar_encoding[ref_subrule_offset] + 1

            return new_stacks

    def _update_state_with_byte(
        self, byte: int, parsing_state: AcceptState
    ) -> AcceptState:
        # suppose we have code point 一, ord('一') = 19968, we need to match 3 bytes
        # we need to match 3 bytes, so we need to call _consume_byte_partial_match 3 times
        return self._update_state_with_bytes(bytes([byte]), parsing_state)

    # @lru_cache(maxsize=32768)
    def _try_accept_bytes(
        self,
        byte_seq: bytes,
        stacks: Set[Tuple[int]],
        partial_utf8: PartialUTF8,
        verbose=True,
    ):
        """
        The difference between accept_bytes and consume_bytes is that accept_bytes returns a boolean and
        consume_bytes returns a new accept state
        """
        if type(byte_seq) is list:
            byte_seq = bytes(byte_seq)
        code_points, new_partial_utf8 = decode_utf8(byte_seq, partial_utf8)
        if verbose:
            logging.debug(
                f"code_points: {code_points}; new_partial_utf8: {new_partial_utf8}"
            )
        new_stacks = self._update_state_with_code_points_for_all_stacks(
            code_points, stacks
        )

        for stack in new_stacks:

            # stack is empty, meaning that the variables are all consumed
            if len(stack) == 0:
                return True
            element_offset = stack[-1]
            if self.partial_utf8_accept_at_element(element_offset, new_partial_utf8):
                return True
        return False

    def _update_state_with_bytes(
        self,
        byte_seq: bytes,
        parsing_state: Optional[AcceptState] = None,
        verbose=True,
    ) -> AcceptState:
        if parsing_state is None:
            parsing_state = self.get_initial_parsing_state()
        stacks = parsing_state.stacks
        partial_utf8 = parsing_state.partial_utf8
        if type(byte_seq) is list:
            byte_seq = bytes(byte_seq)
        code_points, new_partial_utf8 = decode_utf8(byte_seq, partial_utf8)
        if verbose:
            logging.debug(
                f"code_points: {code_points}; new_partial_utf8: {new_partial_utf8}"
            )
        new_stacks = self._update_state_with_code_points_for_all_stacks(
            code_points, stacks
        )

        new_new_stacks = set()
        for stack in new_stacks:
            if len(stack) == 0:
                continue
            element_offset = stack[-1]
            if self.partial_utf8_accept_at_element(element_offset, new_partial_utf8):
                new_new_stacks.add(stack)
        return AcceptState(new_new_stacks, new_partial_utf8)

    ##########################
    #
    # Code point recognition
    #
    ##########################

    @lru_cache(maxsize=30000)
    def _update_state_with_code_point_for_all_stacks(
        self, code_point: int, stacks: Tuple[Tuple[int]]
    ) -> Set[Tuple[int]]:
        """
        consume a character from the stack
        char_code_point: can be a Unicode code point, including ascii code points which are in the range [0, 127]
        """
        new_stacks: Set[Tuple[int]] = set()

        if code_point == 0:
            return new_stacks
        for stack in stacks:
            new_stacks.update(
                self._update_state_with_code_point_for_single_stack(code_point, stack)
            )
        return new_stacks

    @lru_cache(maxsize=30000)
    def _update_state_with_code_point_for_single_stack(
        self, code_point: int, stack: Tuple[int]
    ) -> Set[Tuple[int]]:
        """
        consume a character from the stack
        char_code_point: can be a Unicode code point, including ascii code points which are in the range [0, 127]
        """
        # TODO, the below code will raise an error when the stack is empty, but why is this happening?
        # if len(stacks) == 0:
        #     raise ValueError("Stacks don't contain any stack, meaning that no character can be consumed")
        # code_point = 0 is a special case when the uf8 sequence is not complete, we return an empty stack
        # to indicate that the character is not accepted

        new_stacks: Set[Tuple[int]] = set()
        if code_point == 0:
            return new_stacks
        # stack is empty
        if len(stack) == 0:
            return new_stacks

        element_offset = stack[-1]

        found = self.accept_code_point_at_element(code_point, element_offset)
        if not found:
            return new_stacks

        size = self.grammar_encoding[element_offset]
        element_offset += size + 1
        new_stack = list(stack[:-1])
        if self.grammar_encoding[element_offset]:
            new_stack.append(element_offset)
        # # Explicitly convert list to tuple of int to make it hashable
        new_tuple_stack: Tuple[int, ...] = tuple(new_stack)
        return self.expand_stack_head(new_tuple_stack)

    def _update_state_with_code_points_for_all_stacks(
        self, code_points: List[int], stacks: Set[Tuple[int]], verbose=False
    ) -> Set[Tuple[int]]:
        """
        code points is a list of Unicode code points. For example, the code points for "hello" is [104, 101, 108, 108, 111]
        For unicode string "こんにちは世界", the code points are [12371, 12435, 12395, 12385, 12399, 19990, 30028]

        """
        for i, code_point in enumerate(code_points):
            # for lru_cache to work, we need to convert the list of stacks into a tuple of stacks
            tuple_stacks: Tuple[Tuple[int], ...] = tuple(stacks)
            stacks = self._update_state_with_code_point_for_all_stacks(
                code_point, tuple_stacks
            )
            if len(stacks) > 0 and verbose:
                accepted_code_point = code_points[: i + 1]
                corresponding_char = chr(code_point)
                logging.debug(
                    f"code point {accepted_code_point} corresponding to {corresponding_char} is accepted"
                )
        return stacks

    def _accept_code_points(
        self, code_points: List[int], stacks: Set[Tuple[int]], verbose=False
    ) -> bool:
        stacks = self._update_state_with_code_points_for_all_stacks(
            code_points, stacks, verbose
        )
        return len(stacks) > 0

    @lru_cache(maxsize=30000)
    def accept_code_point_at_element(
        self, code_point: int, element_offset: int
    ) -> bool:
        size = self.grammar_encoding[element_offset]
        # to make idx point to the range_start of the first range
        element_offset += 1
        for i in range(0, size, 2):
            if (
                self.grammar_encoding[element_offset + i]
                <= code_point
                <= self.grammar_encoding[element_offset + i + 1]
            ):
                return True
        return False

    #############################
    #
    # Partial UTF-8 recognition
    #
    #############################

    def partial_utf8_accept_at_element(
        self, element_offset: int, partial_utf8: PartialUTF8
    ) -> bool:
        # Extract the accumulated value and the number of remaining bytes from the partial_utf8 object.
        partial_value = partial_utf8.value
        n_remain = partial_utf8.n_remain

        # Return False if there are no remaining bytes to process or if it's an invalid UTF-8 sequence.
        if n_remain == 1 and partial_value < 2:
            return False

        # If there are no remaining bytes, this means we had already consumed a complete UTF-8 sequence.
        if n_remain <= 0:
            return True

        # Calculate the lowest possible Unicode code point that can be formed with the remaining bytes.
        low = partial_value << (n_remain * 6)
        # Calculate the highest possible Unicode code point by setting all remaining bits to 1.
        high = low | ((1 << (n_remain * 6)) - 1)

        # If the low end of the range is 0 and a specific number of bytes remain, adjust low to the minimum value
        # that can be represented with that number of bytes. This accounts for UTF-8 encoding rules.
        if low == 0:
            if n_remain == 2:
                low = 1 << 11  # Minimum value representable with 2 additional bytes.
            elif n_remain == 3:
                low = 1 << 16  # Minimum value representable with 3 additional bytes.

        # Get the size of the grammar rule starting at the current element_offset.
        size = self.grammar_encoding[element_offset]
        # Move the element_offset to the start of the grammar rule's definition.
        element_offset += 1

        # Iterate over the grammar rule, checking if the range defined by low-high overlaps with any specified ranges.
        for i in range(0, size, 2):
            # If the current range (specified in the grammar encoding) overlaps with the low-high range, return True.
            if intervals_intersect(
                low,
                high,
                self.grammar_encoding[element_offset + i],
                self.grammar_encoding[element_offset + i + 1],
            ):
                return True

        # If no overlap is found with any of the ranges, return False, indicating no valid partial match.
        return False

    #############################
    #
    # String recognition
    #
    #############################

    def _update_state_with_string(self, string: str, parsing_state: AcceptState):
        # _bytes = bytes(string, "utf-8")
        code_points = [ord(char) for char in string]
        stacks = self._update_state_with_code_points_for_all_stacks(
            code_points, parsing_state.stacks
        )
        return AcceptState(stacks, parsing_state.partial_utf8)

    def _accept_prefix(self, string: str, parsing_state: Optional[AcceptState] = None):
        if parsing_state is None:
            parsing_state = self.get_initial_parsing_state()
        new_parsing_state = self._update_state_with_string(string, parsing_state)
        return len(new_parsing_state.stacks) > 0

    def _accept_string(self, string: str, parsing_state: Optional[AcceptState] = None):
        if parsing_state is None:
            parsing_state = self.get_initial_parsing_state()
        new_parsing_state = self._update_state_with_string(string, parsing_state)
        at_least_one_stack_is_empty = any(
            len(stack) == 0 for stack in new_parsing_state.stacks
        )
        return at_least_one_stack_is_empty

    def _can_stop(self, stacks: Set[Tuple[int]]):
        # This happens in practice, but maybe it shouldn't? TODO
        if len(stacks) == 0:
            return True
        # if any of the stack is empty, we can stop
        for stack in stacks:
            if len(stack) == 0:
                return True
        else:
            return False

    def _must_stop(self, stacks: Set[Tuple[int]]):
        return len(stacks) == 0 or all(len(stack) == 0 for stack in stacks)

    #############################
    #
    # Not Used
    #
    #############################

    # For each sub-rule in the grammar, cache whether each byte is accepted.
    @lru_cache(maxsize=None)
    def char_acceptance_at_element(self, element_offset):
        """
        Caches and returns a dictionary indicating whether a Unicode character is accepted
        at a given rule position. This function considers Unicode characters, dynamically
        inserting accepted ranges into a dictionary to optimize memory usage.

        Args:
        - rule_offset: The offset in the grammar encoding where the rule starts.

        Returns:
        - A dictionary where each key is a Unicode character (or range) and the value is True if accepted.
        """
        logging.debug(f"element_offset: {element_offset}")
        acceptance = {}
        num_chars = self.grammar_encoding[element_offset]
        element_offset += 1
        for i in range(0, num_chars, 2):
            start = self.grammar_encoding[element_offset + i]
            end = self.grammar_encoding[element_offset + i + 1]
            for j in range(start, end + 1):
                acceptance[j] = True
        logging.debug(acceptance)
        return acceptance


# backward compatibility, add alias of StringRecognizer to GrammarRecognizer
GrammarRecognizer = StringRecognizer

if __name__ == "__main__":
    # set logging level

    with open("examples/grammars/japanese.ebnf", "r") as file:
        input_text = file.read()
    parsed_grammar = parse_ebnf(input_text)
    logging.debug(f"symbol_ids: \n{parsed_grammar.symbol_table}")

    start_rule_id = parsed_grammar.symbol_table["root"]
    recognizer = StringRecognizer(parsed_grammar.grammar_encoding, start_rule_id)

    japanese = "こんにちは世界"  # "こんにちは世界" doesn't work
    bytes_japanese = bytes(japanese, "utf-8")
    logging.debug(f"bytes_japanese: {bytes_japanese} of length {len(bytes_japanese)}")
    # こんにちは世界

    partial_utf8 = PartialUTF8()

    #######################
    # Japanese
    #######################

    # logging.debug(f"bytes_japanese: {bytes_japanese} of length {len(bytes_japanese)}")
    # head_bytes = bytes_japanese[:8]
    # # partial_utf8 = PartialUTF8()
    # new_stacks, new_partial_utf8 = recognizer._consume_bytes_partial_match(head_bytes, recognizer.stacks, partial_utf8)
    # if len(new_stacks) > 0:
    #     logging.debug("japanese is accepted")
    # else:
    #     logging.debug("japanese is not accepted")

    #######################
    # Now consider the case of progressive matching
    #######################

    byte_tokens = [bytes_japanese[i] for i in range(len(bytes_japanese))]
    # cast into bytes
    byte_tokens = [bytes([byte]) for byte in byte_tokens]

    parsing_state = AcceptState(recognizer.stacks, PartialUTF8())
    for i, byte in enumerate(byte_tokens):
        parsing_state = recognizer._update_state_with_bytes(byte, parsing_state)
        logging.debug(f"new partial utf8: {parsing_state.partial_utf8}")
        if len(parsing_state.stacks) > 0:
            logging.debug(f"byte {byte} is accepted")
        else:
            logging.debug(f"byte {byte} is not accepted")

    # korean = "안녕하세요"
    # korean_bytes = bytes(korean, "utf-8")
    # head_bytes = korean_bytes[:8]
    # logging.debug(f"korean_bytes: {korean_bytes} of length {len(korean_bytes)}")
    # new_stacks, new_partial_utf8 = recognizer._consume_bytes_partial_match(head_bytes, recognizer.stacks, partial_utf8)
    # if len(new_stacks) > 0:
    #     logging.debug("korean is accepted")
    # else:
    #     logging.debug("korean is not accepted")

    # res = recognizer._accept_string("12222", recognizer.stacks)
    # logging.debug(f"12222: {res}")
    # res = recognizer._accept_string("12222+", recognizer.stacks)
    # logging.debug(f"12222+: {res}")



#################
# DATA STRUCTURES
#################

import logging
from typing import List

from transformers_cfg.tokenization.tokenizer import TCFG_Tokenizer

logger = logging.getLogger(__name__)

LEAF = -1


class TokenTrie:
    # Not used anymore, replaced by ByteTrie
    def __init__(self, hf_tokenizer):
        self.eos_token_id = hf_tokenizer.eos_token_id
        self.tcfg_tokenizer = TCFG_Tokenizer.from_hf_tokenizer(hf_tokenizer)
        self.trie = {}

        self.load_token_as_bytes()

    def id2str(self, token_id):
        return self.tokens[token_id]

    def __len__(self):
        return self.tcfg_tokenizer.real_vocab_size()

    def load_token_as_bytes(self):
        tokens: List[bytes] = self.tcfg_tokenizer.get_tokens_as_bytes()
        for token_id, token_bytes in enumerate(tokens):
            if token_bytes is not None:
                self.insert_into_trie(self.trie, token_bytes, token_id)

    def insert_into_trie(self, trie, token_bytes, token_id):
        current = trie
        for byte in token_bytes:
            if byte not in current:
                current[byte] = {}
            current = current[byte]
        current[LEAF] = token_id


if __name__ == "__main__":
    from transformers import AutoTokenizer

    tokenizer = AutoTokenizer.from_pretrained("gpt2")
    token_trie = TokenTrie(tokenizer)



from transformers import AutoTokenizer


from typing import Dict, List


class ByteProxyMapping:
    def __init__(self, tokenizer):
        # check if the tokenizer is fast, if so, convert it to slow
        if tokenizer.is_fast:
            tokenizer = AutoTokenizer.from_pretrained(
                tokenizer.name_or_path, use_fast=False
            )
        self.tokenizer = tokenizer

        # if tokenizer doesn't have byte_encoder(which is the case for llama-3), use gpt2_tokenizer
        if not hasattr(tokenizer, "byte_encoder"):
            gpt2_tokenizer = AutoTokenizer.from_pretrained("gpt2", use_fast=False)
            tokenizer.byte_encoder = gpt2_tokenizer.byte_encoder
            tokenizer.byte_decoder = gpt2_tokenizer.byte_decoder

        self.byte2proxychar: Dict[int, str] = tokenizer.byte_encoder
        self.proxychar2byte: Dict[str, int] = tokenizer.byte_decoder

        # code point to byte
        self.cdp2byte: Dict[int, int] = {
            ord(c): b for c, b in self.proxychar2byte.items()
        }
        self.byte2cdp: Dict[int, int] = {v: k for k, v in self.cdp2byte.items()}
        self.PROXY_CDP_SET = set(self.cdp2byte.keys())
        # [33, 126] and [161,172, [174, 323], in total 94 + 12 + 150 = 256(N.B. 173 is a control character)

    def encode_byte2proxychar_cdp(self, byte: int) -> int:
        assert 0 <= byte < 256, f"byte: {byte} is not in the range [0, 256)"
        return ord(self.byte2proxychar[byte])

    def decode_proxychar2byte_cdp(self, cdp: int) -> int:
        byte_int: int = self.cdp2byte[cdp]
        assert 0 <= byte_int < 256, f"byte: {byte_int} is not in the range [0, 256)"
        return byte_int

    def decode_proxytoken2bytes(self, proxy_token: str) -> bytes:
        bytes_seq: List[int] = [
            self.decode_proxychar2byte_cdp(ord(c)) for c in proxy_token
        ]
        return bytes(bytes_seq)

    def map(self, proxy_token: str) -> bytes:
        return self.decode_proxytoken2bytes(proxy_token)

    def token2bytes(self, token: str) -> bytes:
        bytes_seq: List[int] = [self.proxychar2byte[c] for c in token]
        return bytes(bytes_seq)


class LLAMAByteProxyMapper:
    def __init__(self):
        pass

    def map(self, proxy_token: str) -> bytes:
        return self.decode_proxytoken2bytes(proxy_token)

    def decode_proxytoken2bytes(self, proxy_token: str) -> bytes:
        if proxy_token.startswith("<0x"):
            hex_value: str = proxy_token[3:-1]
            return bytes.fromhex(hex_value)
        else:
            # ad hoc fix for BPE
            if proxy_token.startswith("▁"):
                proxy_token = proxy_token.replace("▁", " ")
            return proxy_token.encode("utf-8")


if __name__ == "__main__":

    gpt2_tokenizer = AutoTokenizer.from_pretrained("gpt2")

    # gpt2_tokenizer.encode("´")

    byteproxymapper = ByteProxyMapping(gpt2_tokenizer)

    for i in range(256):
        print(f"{i}: {byteproxymapper.encode_byte2proxychar_cdp(i)}")

    # decode a byte

    byte = 162
    print(f"proxy code point set: {byteproxymapper.PROXY_CDP_SET}")
    print(f"len(proxy code point set): {len(byteproxymapper.PROXY_CDP_SET)}")
    for i in range(33, 127):
        if i not in byteproxymapper.PROXY_CDP_SET:
            print(f"{i} not in proxy code point set")
    for i in range(161, 324):
        if i not in byteproxymapper.PROXY_CDP_SET:
            print(f"{i} not in proxy code point set")



import argparse
import logging
import sys
from typing import List

logger = logging.getLogger(__name__)

END_OF_ALTERNATE_MARKER = 0
END_OF_RULE_MARKER = 0
END_OF_GRAMMAR_MARKER = 0xFFFF
TO_BE_FILLED_MARKER = 0
REF_RULE_MARKER = 1
LITERAL_MARKER = 2


########################
# EBNF Grammar Parsing #
########################


class ParseState:
    def __init__(self):
        self.symbol_table = {}
        self.grammar_encoding = []  # old name: out_grammar

    def print(self, file=sys.stdout):
        print_grammar(file, self)


def get_symbol_id(state: ParseState, symbol_name: str) -> int:
    if symbol_name not in state.symbol_table:
        state.symbol_table[symbol_name] = len(state.symbol_table)
    return state.symbol_table[symbol_name]


def generate_symbol_id(state: ParseState, base_name: str) -> int:
    next_id = len(state.symbol_table)
    state.symbol_table[base_name + "_" + str(next_id)] = next_id
    return next_id


def is_word_char(c: str) -> bool:
    """
    Check if a char is  a-z, A-Z, 0-9, -, _, i.e., chars allowed as rule names
    Returns:

    """
    return c.isalnum() or c == "-" or c == "_"


def hex_to_int(c: str) -> int:
    """
    Convert a hex char to int, c should be in the range of 0-9, a-f, A-F
    case insensitive
    Args:
        c:  a hex char
    Returns:
        int: the int value of the hex char
    """
    if c.isdigit():
        return int(c)
    elif "a" <= c.lower() <= "f":
        return ord(c.lower()) - ord("a") + 10
    return -1


def remove_leading_white_space(src, rm_leading_newline):
    """
    Skips over whitespace and comments in the input string.

    This function processes the input string, skipping over any spaces, tabs,
    and content following a '#' character, which denotes a comment. The parsing
    of a comment continues until the end of the line (denoted by newline characters
    '\r' or '\n'). If the 'rm_leading_newline' parameter is set to False, the function
    will stop processing and return the remaining string upon encountering a
    newline character, otherwise it will skip over newline characters as well.

    Parameters:
    src (str): The input string to be processed.
    rm_leading_newline (bool): A flag indicating whether encountering a newline character
                       should stop the parsing (False) or if it should be skipped (True).

    Returns:
    str: The remaining portion of the input string after skipping whitespace and comments.
    """
    pos = 0
    while pos < len(src) and (src[pos].isspace() or src[pos] == "#"):
        if src[pos] == "#":
            while pos < len(src) and src[pos] not in ("\r", "\n"):
                pos += 1
        else:
            if not rm_leading_newline and src[pos] in ("\r", "\n"):
                break
            pos += 1
    return src[pos:]


def parse_name(src) -> (str, str):
    """
    parse the leading name from the input string
    Args:
        src:  the input grammar string

    Returns:
        name, remaining_src
    """
    pos = 0
    while pos < len(src) and is_word_char(src[pos]):
        pos += 1
    if pos == 0:
        raise RuntimeError("expecting name at " + src)
    return src[:pos], src[pos:]


def parse_char(src) -> (str, str):
    """
    parse the leading char from the input string
    :param src:
    :return: char, remaining_src
    """

    # if we have a backslash, it's maybe an escape
    if src[0] == "\\":
        esc = src[1]
        if esc == "x":
            first = hex_to_int(src[2])
            if first > -1:
                second = hex_to_int(src[3])
                if second > -1:
                    return chr((first << 4) + second), src[4:]
            raise RuntimeError("expecting \\xNN at " + src)
        elif esc == "u":
            if len(src) >= 6:
                hex_value = src[2:6]
                if all(c in "0123456789ABCDEFabcdef" for c in hex_value):
                    return chr(int(hex_value, 16)), src[6:]
                raise RuntimeError("expecting \\uXXXX at " + src)
            raise RuntimeError("incomplete \\uXXXX escape at " + src)
        elif esc == "U":
            if len(src) >= 10:
                hex_value = src[2:10]
                if all(c in "0123456789ABCDEFabcdef" for c in hex_value):
                    return chr(int(hex_value, 16)), src[10:]
                raise RuntimeError("expecting \\UXXXXXXXX at " + src)
            raise RuntimeError("incomplete \\UXXXXXXXX escape at " + src)
        elif esc in ('"', "[", "]"):
            return esc, src[2:]
        elif esc == "r":
            return "\r", src[2:]
        elif esc == "n":
            return "\n", src[2:]
        elif esc == "t":
            return "\t", src[2:]
        elif esc == "\\":
            return "\\", src[2:]
        raise RuntimeError("unknown escape at " + src)
    elif src:
        return src[0], src[1:]
    raise RuntimeError("unexpected end of input")


def _parse_rhs_literal_string(src: str, outbuf: List[int]) -> str:
    assert src[0] == '"', f"rule should start with '\"', but got {src[0]}"
    remaining_src = src[1:]

    # advance until we get an end quote or run out of input
    while remaining_src and remaining_src[0] != '"':
        char, remaining_src = parse_char(remaining_src)
        outbuf.append(LITERAL_MARKER)
        outbuf.append(ord(char))
        outbuf.append(ord(char))

    # in case we ran out of input before finding the end quote
    if not remaining_src:
        raise RuntimeError(f"expecting an end quote at {src},but not found")

    # remove the end quote and return the remaining string
    return remaining_src[1:]


def _parse_rhs_char_ranges(src: str, outbuf: List[int]) -> str:
    assert src[0] == "[", f"rule should start with '[', but got {src[0]}"
    remaining_src = src[1:]
    start_idx = len(outbuf)
    # num chars in range - replaced at end of loop
    outbuf.append(TO_BE_FILLED_MARKER)
    while remaining_src and remaining_src[0] != "]":
        char, remaining_src = parse_char(remaining_src)

        outbuf.append(ord(char))
        if remaining_src[0] == "-" and remaining_src[1] != "]":
            endchar_pair, remaining_src = parse_char(remaining_src[1:])
            outbuf.append(ord(endchar_pair))
        else:
            # This is the case for enumerate, e.g., [0123456789], [abcdef]
            # Each char is considered as a range of itself, i.e., c-c
            outbuf.append(ord(char))
    if not remaining_src:
        raise RuntimeError(
            f"expecting an ] at {src},but not found, is the char range closed?"
        )
    # replace num chars with actual
    outbuf[start_idx] = len(outbuf) - start_idx - 1
    return remaining_src[1:]


def _parse_rhs_symbol_reference(src: str, state: ParseState, outbuf: List[int]) -> str:
    assert is_word_char(src[0]), f"rule should start with a word char, but got {src[0]}"
    name, remaining_src = parse_name(src)
    ref_rule_id = get_symbol_id(state, name)
    outbuf.append(REF_RULE_MARKER)
    outbuf.append(ref_rule_id)
    return remaining_src


def _parse_rhs_grouping(
    remaining_src: str, state: ParseState, rule_name: str, outbuf: List[int]
) -> str:
    assert (
        remaining_src[0] == "("
    ), f"rule should start with '(', but got {remaining_src[0]}"
    remaining_src = remove_leading_white_space(remaining_src[1:], True)
    # parse nested alternates into synthesized rule
    synthetic_rule_id = generate_symbol_id(state, rule_name)
    remaining_src = parse_rhs(state, remaining_src, rule_name, synthetic_rule_id, True)
    # output reference to synthesized rule
    outbuf.append(REF_RULE_MARKER)
    outbuf.append(synthetic_rule_id)

    if not remaining_src or remaining_src[0] != ")":
        raise RuntimeError("expecting ')' at " + remaining_src)
    return remaining_src[1:]


def _parse_rhs_repetition_operators(
    remaining_src: str,
    state: ParseState,
    rule_name: str,
    last_sym_start: int,
    outbuf: List[int],
) -> str:
    assert remaining_src[0] in (
        "*",
        "+",
        "?",
    ), f"rule should start with '*', '+', or '?', but got {remaining_src[0]}"
    out_grammar = state.grammar_encoding
    # last_sym_start = len(outbuf)

    # apply transformation to previous symbol (last_sym_start -
    # end) according to rewrite rules:
    # S* --> S' ::= S S' |
    # S+ --> S' ::= S S' | S
    # S? --> S' ::= S |
    sub_rule_id = generate_symbol_id(state, rule_name)
    out_grammar.append(sub_rule_id)
    sub_rule_offset = len(out_grammar)
    # placeholder for size of 1st alternate
    out_grammar.append(TO_BE_FILLED_MARKER)
    # add preceding symbol to generated rule
    out_grammar.extend(outbuf[last_sym_start:])
    if remaining_src[0] in ("*", "+"):
        # cause generated rule to recurse
        out_grammar.append(REF_RULE_MARKER)
        out_grammar.append(sub_rule_id)
    # apply actual size
    out_grammar[sub_rule_offset] = len(out_grammar) - sub_rule_offset
    # mark end of 1st alternate
    out_grammar.append(END_OF_ALTERNATE_MARKER)
    sub_rule_offset = len(out_grammar)
    # placeholder for size of 2nd alternate
    out_grammar.append(TO_BE_FILLED_MARKER)
    if remaining_src[0] == "+":
        # add preceding symbol as alternate only for '+'
        out_grammar.extend(outbuf[last_sym_start:])
    # apply actual size of 2nd alternate
    out_grammar[sub_rule_offset] = len(out_grammar) - sub_rule_offset
    # mark end of 2nd alternate, then end of rule
    out_grammar.append(END_OF_ALTERNATE_MARKER)
    out_grammar.append(END_OF_RULE_MARKER)

    # in original rule, replace previous symbol with reference to generated rule
    outbuf[last_sym_start:] = [REF_RULE_MARKER, sub_rule_id]
    return remaining_src[1:]


def parse_simple_rhs(state, rhs: str, rule_name: str, outbuf, is_nested):
    simple_rhs_offset = len(outbuf)

    # sequence size, will be replaced at end when known
    outbuf.append(TO_BE_FILLED_MARKER)

    last_sym_start = len(outbuf)
    remaining_rhs = rhs
    while remaining_rhs:
        if remaining_rhs[0] == '"':  # literal string
            # mark the start of the last symbol, for repetition operator
            last_sym_start = len(outbuf)
            remaining_rhs = _parse_rhs_literal_string(remaining_rhs, outbuf)
        elif remaining_rhs[0] == "[":  # char range(s)
            # mark the start of the last symbol, for repetition operator
            last_sym_start = len(outbuf)
            remaining_rhs = _parse_rhs_char_ranges(remaining_rhs, outbuf)
        elif is_word_char(remaining_rhs[0]):  # rule reference
            # mark the start of the last symbol, for repetition operator
            last_sym_start = len(outbuf)
            remaining_rhs = _parse_rhs_symbol_reference(remaining_rhs, state, outbuf)
        elif remaining_rhs[0] == "(":  # grouping
            # mark the start of the last symbol, for repetition operator
            last_sym_start = len(outbuf)
            remaining_rhs = _parse_rhs_grouping(remaining_rhs, state, rule_name, outbuf)
        elif remaining_rhs[0] in ("*", "+", "?"):  # repetition operator
            # No need to mark the start of the last symbol, because we already did it
            if len(outbuf) - simple_rhs_offset - 1 == 0:
                raise RuntimeError(
                    "expecting preceeding item to */+/? at " + remaining_rhs
                )
            remaining_rhs = _parse_rhs_repetition_operators(
                remaining_rhs, state, rule_name, last_sym_start, outbuf
            )
        else:
            # case for newline, i.e., end of rule
            assert remaining_rhs[0] in [
                "\n",
                "|",
                ")",
            ], f"rule should end with newline or '|', but got {remaining_rhs[0]}"
            # we break here so that we call parse_rule again to parse the next rule
            break
        # Here we do not rm newline deliberately so that we know the rhs is ended
        remaining_rhs = remove_leading_white_space(
            remaining_rhs, rm_leading_newline=is_nested
        )

    # apply actual size of this alternate sequence
    outbuf[simple_rhs_offset] = len(outbuf) - simple_rhs_offset
    # mark end of alternate
    outbuf.append(END_OF_ALTERNATE_MARKER)
    return remaining_rhs


def parse_rhs(state, rhs: str, rule_name, rule_id, is_nested):
    outbuf = []
    remaining_rhs = parse_simple_rhs(state, rhs, rule_name, outbuf, is_nested)
    while remaining_rhs and remaining_rhs[0] == "|":
        remaining_rhs = remove_leading_white_space(remaining_rhs[1:], True)
        remaining_rhs = parse_simple_rhs(
            state, remaining_rhs, rule_name, outbuf, is_nested
        )

    # Now we have finished parsing the rhs, we can add the rule to the grammar_encoding
    state.grammar_encoding.append(rule_id)
    state.grammar_encoding.extend(outbuf)
    state.grammar_encoding.append(END_OF_RULE_MARKER)
    return remaining_rhs


def parse_rule(state: ParseState, rule_text: str) -> str:
    name, remaining_rule_text = parse_name(rule_text)
    remaining_rule_text = remove_leading_white_space(remaining_rule_text, False)
    # check if the rule is already defined, TODO: what will happen if the rule is already defined?
    rule_id = get_symbol_id(state, name)

    if remaining_rule_text[:3] != "::=":
        raise RuntimeError("expecting ::= at " + remaining_rule_text)
    remaining_rule_text = remove_leading_white_space(remaining_rule_text[3:], True)

    remaining_rule_text = parse_rhs(state, remaining_rule_text, name, rule_id, False)

    if remaining_rule_text and remaining_rule_text[0] == "\r":
        remaining_rule_text = (
            remaining_rule_text[2:]
            if remaining_rule_text[1] == "\n"
            else remaining_rule_text[1:]
        )
    elif remaining_rule_text and remaining_rule_text[0] == "\n":
        remaining_rule_text = remaining_rule_text[1:]
    elif remaining_rule_text:
        raise RuntimeError("expecting newline or end at " + remaining_rule_text)
    return remove_leading_white_space(remaining_rule_text, True)


def parse_ebnf(grammar_text: str) -> ParseState:
    try:
        state = ParseState()
        remaining_grammar_text = remove_leading_white_space(grammar_text, True)
        last_grammar_repr = ""
        while remaining_grammar_text:
            if last_grammar_repr:
                last_parsed_rule_len = len(last_grammar_repr) - len(
                    remaining_grammar_text
                )
                logger.debug(
                    f"last_parsed_rule: {last_grammar_repr[:last_parsed_rule_len]}"
                )
            last_grammar_repr = remaining_grammar_text
            remaining_grammar_text = parse_rule(state, remaining_grammar_text)
        state.grammar_encoding.append(END_OF_GRAMMAR_MARKER)
        return state
    except RuntimeError as err:
        logger.warning("error parsing grammar:", err)
        return ParseState()


###################################
# EBNF Grammar Parsing ends here  #
###################################


def break_grammar_into_rules(grammar_encoding: List[int]) -> List[List[int]]:
    offset = 0
    # we loop until we reach the end of the grammar_encoding
    rule_encodings = []
    i = 0
    while i < len(grammar_encoding) - 2:
        if (
            grammar_encoding[i] == END_OF_ALTERNATE_MARKER
            and grammar_encoding[i + 1] == END_OF_RULE_MARKER
        ):
            rule_encodings.append(grammar_encoding[offset : i + 2])
            offset = i + 2
            # skip the END_OF_RULE_MARKER
            # This is mandatory because if we do not skip the END_OF_RULE_MARKER
            # we fail in the case where the next rule has rule_id 0
            i += 1
        i += 1
    return rule_encodings


def break_rule_into_elements(rule_encoding: List[int]) -> List[List[int]]:
    rule_id = rule_encoding.pop(0)
    end_of_rule_marker = rule_encoding.pop(-1)
    assert (
        end_of_rule_marker == END_OF_RULE_MARKER
    ), f"rule should end with {END_OF_RULE_MARKER}, but got {end_of_rule_marker}"

    offset = 0
    elements = []
    while offset < len(rule_encoding):
        element_size = rule_encoding[offset]
        assert (
            rule_encoding[offset + element_size] == END_OF_ALTERNATE_MARKER
        ), f"element should end with {END_OF_ALTERNATE_MARKER}, but got {rule_encoding[offset + element_size]}"
        elements.append(rule_encoding[offset : offset + element_size + 1])
        offset += element_size + 1
    return elements


def _print_annotated_grammar(file, grammar_encoding, symbol_id_names, index=0):
    rule_id = grammar_encoding[index]
    print(f"<{index}>{symbol_id_names[rule_id]} ::=", end=" ", file=file)
    pos = index + 1
    while grammar_encoding[pos]:
        if pos - 1 > index:
            print("|", end=" ", file=file)
        pos += 1  # sequence size, not needed here
        while grammar_encoding[pos]:
            if grammar_encoding[pos] == REF_RULE_MARKER:
                ref_rule_id = grammar_encoding[pos + 1]
                print(
                    f"<{pos}>{symbol_id_names[ref_rule_id]}",
                    end=" ",
                    file=file,
                )
                pos += 2
            else:
                print("<{}>[".format(pos), end="", file=file)
                num_chars = grammar_encoding[pos]
                pos += 1

                for i in range(0, num_chars, 2):
                    print(
                        "{}-".format(chr(grammar_encoding[pos + i])), end="", file=file
                    )
                    if i + 1 < num_chars:
                        print(
                            "{}".format(chr(grammar_encoding[pos + i + 1])),
                            end="",
                            file=file,
                        )
                print("]", end=" ", file=file)
                pos += num_chars
        pos += 1
    print(file=file)
    return pos + 1


def print_grammar(file, state):
    pos = 0
    symbol_id_names = {v: k for k, v in state.symbol_table.items()}
    print("Grammar Rules:", file=file)
    while (
        pos < len(state.grammar_encoding)
        and state.grammar_encoding[pos] != END_OF_GRAMMAR_MARKER
    ):
        pos = _print_annotated_grammar(
            file, state.grammar_encoding, symbol_id_names, pos
        )
    if pos > len(state.grammar_encoding):
        raise Warning(f"grammar_encoding is not ended with {END_OF_GRAMMAR_MARKER}")
    pos = 0
    print("\nGrammar Hex representation:", file=file)
    while (
        pos < len(state.grammar_encoding)
        and state.grammar_encoding[pos] != END_OF_GRAMMAR_MARKER
    ):
        print(f"{state.grammar_encoding[pos]:04x}", end=" ", file=file)
        pos += 1
    if pos > len(state.grammar_encoding):
        raise Warning(f"grammar_encoding is not ended with {END_OF_GRAMMAR_MARKER}")
    else:
        print("ffff\n")

    print("Rules Decimal representation:", file=file)
    # we loop until we reach the end of the grammar_encoding
    rule_encodings = break_grammar_into_rules(state.grammar_encoding)
    for rule_encoding in rule_encodings:
        rule_id = rule_encoding[0]
        print(
            f"<{rule_id}> {break_rule_into_elements(rule_encoding)}",
            file=file,
        )


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Parse EBNF grammar files.")
    parser.add_argument(
        "-g",
        "--grammar-file",
        nargs="?",
        default="examples/grammars/json.ebnf",
        help="Path to the grammar file (default: examples/grammars/json.ebnf)",
    )

    args = parser.parse_args()

    with open(args.grammar_file, "r") as file:
        input_text = file.read()
    parsed_grammar = parse_ebnf(input_text)
    parsed_grammar.print()
    print(f"symbol_ids: \n{parsed_grammar.symbol_table}")

    start_rule_id = parsed_grammar.symbol_table["root"]



import unittest

from transformers import LlamaTokenizerFast

from tests.test_accept_token_sequence._test_accept_tokens_mixin import (
    TokenizerTesterMixin,
)


class LlamaTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = LlamaTokenizerFast
    pretrained_name = "saibo/llama-1B"

    def setUp(self):
        super().setUp()



import unittest

from transformers import T5TokenizerFast

from tests.test_accept_token_sequence._test_accept_tokens_mixin import (
    TokenizerTesterMixin,
)


# @unittest.skip("T5Tokenizer's mapping is not well defined, not working")
class T5TokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = T5TokenizerFast
    pretrained_name = "t5-small"

    def setUp(self):
        super().setUp()


class TestT5TokenizerUnkToken(unittest.TestCase):
    def test_unk_token(self):
        tokenizer = T5TokenizerFast.from_pretrained("t5-small")

        unk_token_id = tokenizer.unk_token_id
        unk_token = tokenizer.unk_token

        # open curly brace is an unk token
        curly_brace_open = "{"
        # we take the 2nd token because the first token is the space token
        curly_brace_open_id = tokenizer.encode(curly_brace_open)[1]
        self.assertEqual(curly_brace_open_id, unk_token_id)

        curly_brace_close = "}"
        curly_brace_close_id = tokenizer.encode(curly_brace_close)[1]
        self.assertEqual(curly_brace_close_id, unk_token_id)

        eos_token_id = tokenizer.eos_token_id
        # tab in t5 signifies the end of a line
        tab = "\t"
        tab_id = tokenizer.encode(tab)[0]
        self.assertEqual(tab_id, eos_token_id)

        # newline in t5 signifies the end of a line
        newline = "\n"
        newline_id = tokenizer.encode(newline)[0]
        self.assertEqual(newline_id, eos_token_id)



import unittest

from transformers import GPT2TokenizerFast

from tests.test_accept_token_sequence._test_accept_tokens_mixin import (
    TokenizerTesterMixin,
)


class GPT2TokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = GPT2TokenizerFast
    pretrained_name = "gpt2"

    def setUp(self):
        super().setUp()



from unittest import TestCase

from transformers_cfg.parser import parse_ebnf
from transformers_cfg.recognizer import StringRecognizer
from dataclasses import dataclass


@dataclass
class OvernightTestCase:
    name: str
    overnight: str


valid_overnight_sentences = [
    OvernightTestCase(
        "simple_request", "(listValue (getProperty en.block.block1 width))"
    ),
    OvernightTestCase(
        "simple_filter",
        "(listValue (filter (getProperty (singleton en.block) !type) height = 3 en.inch))",
    ),
    OvernightTestCase(
        "count_values",
        "(listValue (countComparative (getProperty (singleton en.block) !type) shape >= 2))",
    ),
    OvernightTestCase(
        "ensure_property",
        "(listValue (filter (getProperty (singleton en.block) !type) (ensureNumericProperty width) <= (ensureNumericEntity 3 en.inch)))",
    ),
    OvernightTestCase(
        "above",
        "(listValue (filter (filter (getProperty (singleton en.block) !type) (reverse above) = en.block.block1) above = en.block.block1))",
    ),
    OvernightTestCase(
        "reverse right",
        "(listValue (filter (getProperty (singleton en.block) !type) right = (filter (getProperty (singleton en.block) !type) (reverse right) = en.block.block1)))",
    ),
    OvernightTestCase(
        "agg",
        "(listValue (superlative (getProperty (singleton en.block) !type) max (ensureNumericProperty length)))",
    ),
    OvernightTestCase(
        "nested_filters",
        "(listValue (filter (filter (getProperty (singleton en.block) !type) (reverse above) = en.block.block1) (reverse right) = en.block.block1))",
    ),
    OvernightTestCase(
        "shape",
        "(listValue (filter (getProperty (singleton en.block) !type) shape != en.shape.pyramid))",
    ),
    OvernightTestCase(
        "is_special",
        "(listValue (filter (filter (getProperty (singleton en.block) !type) is_special) left = en.block.block1))",
    ),
    OvernightTestCase(
        "two_blocks",
        "(listValue (filter (getProperty (singleton en.block) !type) left = (concat en.block.block1 en.block.block2)))",
    ),
    OvernightTestCase(
        "count_superlative",
        "(listValue (countSuperlative (getProperty (singleton en.block) !type) min (reverse above) (getProperty (singleton en.block) !type)))",
    ),
    OvernightTestCase(
        "long_query",
        "(listValue (filter (getProperty (singleton en.block) !type) (ensureNumericProperty height) > (ensureNumericEntity (getProperty en.block.block1 height))))",
    ),
    OvernightTestCase(
        "2_value",
        "(listValue (countComparative (getProperty (singleton en.block) !type) left < 2 (getProperty (singleton en.block) !type)))",
    ),
    OvernightTestCase(
        "concat_shapes",
        "(listValue (filter (getProperty (singleton en.block) !type) shape = (concat en.shape.pyramid en.shape.cube)))",
    ),
]


valid_overnight_prefixes = [
    OvernightTestCase("empty_string", ""),
    OvernightTestCase(
        "unbalanced_paranthesis", "(listValue (getProperty en.block.block1 width"
    ),
    OvernightTestCase("undefined_argument", "(listValue (getProperty en.block.block1"),
    OvernightTestCase(
        "left_comarisson",
        "(listValue (filter (getProperty (singleton en.block) !type) (ensureNumericProperty length) >=",
    ),
]

invalid_overnight_sentences = [
    OvernightTestCase(
        "unknown_property", "(listValue (getProperty en.block.block1 sparkliness))"
    ),
    OvernightTestCase("property", "(getProperty en.block.block1 width)"),
    OvernightTestCase("number_value", "3 en.inch"),
    OvernightTestCase(
        "extra_space", "(listValue ( getProperty en.block.block1 width))"
    ),
    OvernightTestCase("empty_operator", "(listValue (getProperty ))"),
    OvernightTestCase("empty_paranthesis", "()"),
    OvernightTestCase("missing_argument", "(listValue (getProperty en.block.block1 ))"),
    OvernightTestCase(
        "inexisting_shape",
        "(listValue (filter (getProperty (singleton en.block) !type) shape = (concat en.shape.pyramid en.shape.sphere)))",
    ),
]


class Test_parsing_overnight_object(TestCase):
    def setUp(self):
        with open(f"examples/grammars/overnight.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)
        print("PARSED GRAMMAR:", parsed_grammar.grammar_encoding, flush=True)
        start_rule_id = parsed_grammar.symbol_table["root"]
        self.recognizer = StringRecognizer(
            parsed_grammar.grammar_encoding, start_rule_id
        )

    def test_valid_sentence(self):

        for overnight_test_case in valid_overnight_sentences:
            self.assertEqual(
                True,
                self.recognizer._accept_string(overnight_test_case.overnight),
                msg=f"Failed on {overnight_test_case.name}, {overnight_test_case.overnight}",
            )
        for overnight_test_case in (
            valid_overnight_prefixes + invalid_overnight_sentences
        ):
            self.assertEqual(
                False,
                self.recognizer._accept_string(overnight_test_case.overnight),
                msg=f"Failed on {overnight_test_case.name}, {overnight_test_case.overnight}",
            )

    def test_valid_prefixes(self):
        for overnight_test_case in valid_overnight_sentences + valid_overnight_prefixes:
            self.assertEqual(
                True,
                self.recognizer._accept_prefix(overnight_test_case.overnight),
                msg=f"Failed on {overnight_test_case.name}, {overnight_test_case.overnight}",
            )

        for overnight_test_case in invalid_overnight_sentences:
            self.assertEqual(
                False,
                self.recognizer._accept_prefix(overnight_test_case.overnight),
                msg=f"Failed on {overnight_test_case.name}, {overnight_test_case.overnight}",
            )



import logging
from functools import lru_cache
from typing import Dict, List, Set, Tuple
from collections import deque

from transformers_cfg.tokenization.middle.TokenizerMiddleMapping import (
    TokenizerMiddleMapping,
)
from transformers_cfg.tokenization.tokenizer import TCFG_Tokenizer

logger = logging.getLogger(__name__)


class TrieNode:
    def __init__(self):
        self.children = {}
        self.is_end_of_word = False
        self.token_id = None


class ByteTrie:
    def __init__(self):
        self.root = TrieNode()

    def insert(self, word, token_id=None):
        node = self.root
        for char in word:
            if char not in node.children:
                node.children[char] = TrieNode()
            node = node.children[char]
        node.is_end_of_word = True
        node.token_id = token_id

    def search(self, word):
        node = self.root
        for char in word:
            if char not in node.children:
                return False
            node = node.children[char]
        return node.is_end_of_word

    def start_with_prefix(self, prefix):
        node = self.root
        for char in prefix:
            if char not in node.children:
                return False
            node = node.children[char]
        return True

    @classmethod
    def from_tokenizer(cls, tokenizer):
        vocab: Dict[str, int] = tokenizer.get_vocab()
        trie = cls()
        mapping = TokenizerMiddleMapping.from_hf_tokenizer(tokenizer)
        TCFG_tokenizer = TCFG_Tokenizer.from_hf_tokenizer(tokenizer)

        token_ids_to_ignore: Set[
            int
        ] = TCFG_tokenizer.get_special_token_ids_to_excluded()
        for token_id in range(TCFG_tokenizer.real_vocab_size()):
            if token_id not in token_ids_to_ignore:
                byte_repr = mapping.map(token_id)
                trie.insert(byte_repr, token_id)
        trie.vocab_size = len(vocab)
        return trie

    @lru_cache(maxsize=128)
    def __len__(self):
        # return len(self.dfs(verbose=False))
        return self.vocab_size

    def dfs(self, accept=lambda x: True, verbose=False) -> List[Tuple[List[int], int]]:
        result = []
        counter = {"visited": 0, "pruned": 0}
        _dfs(self.root, [], result, accept, counter)
        return result

    def bfs(
        self, predicate=lambda x: True, verbose=False
    ) -> List[Tuple[List[int], int]]:
        queue = deque([(self.root, [])])
        valid_byte_seqs: List[Tuple[List[int], int]] = []
        counter = {"visited": 0, "pruned": 0}

        while queue:
            counter["visited"] += 1
            node, byte_seq = queue.popleft()
            if predicate(byte_seq):
                if node.is_end_of_word:
                    valid_byte_seqs.append((byte_seq, node.token_id))
                for char, next_node in node.children.items():
                    new_byte_seq: List[int] = byte_seq.copy()
                    new_byte_seq.append(char)
                    queue.append((next_node, new_byte_seq))
            else:
                counter["pruned"] += 1
        return valid_byte_seqs

    def get_next_token_acceptance(
        self, accept=lambda x: True, accept_eos=True, eos_token_id=None
    ) -> List[bool]:
        valid_byte_seqs: List[Tuple[List[int], int]] = self.bfs(accept, verbose=True)
        valid_token_ids: List[int] = [token_id for _, token_id in valid_byte_seqs]
        token_acceptance: List[bool] = [False] * (len(self))

        for token_id in valid_token_ids:
            token_acceptance[token_id] = True
        if not accept_eos:
            # eos_token is mapped to an empty string, so it's always accepted regardless of the accept function
            # this can be undesirable, so we can set it to False to ignore it
            token_acceptance[eos_token_id] = False
        return token_acceptance

    def visualize(self, max_depth=3):
        def _visualize(node, prefix, depth):
            if depth > max_depth:
                return
            for char, next_node in node.children.items():
                print(f"{prefix}{char} (Token ID: {next_node.token_id})")
                _visualize(next_node, prefix + "  ", depth + 1)

        print("Visualizing ByteTrie:")
        _visualize(self.root, "", 1)


def _dfs(
    node,
    cur_byte_seq: List[int],
    result: List[Tuple[List[int], int]],
    accept: callable,
    counter: Dict[str, int],
):
    counter["visited"] += 1
    if accept(cur_byte_seq):
        if node.is_end_of_word:
            result.append((cur_byte_seq, node.token_id))
        for char, next_node in node.children.items():
            new_byte_seq: List[int] = cur_byte_seq.copy()
            new_byte_seq.append(char)
            _dfs(next_node, new_byte_seq, result, accept, counter)
    else:
        # Skip the entire subtree if the predict function returns False
        counter["pruned"] += 1
        return


def starts_with_prefix(prefix, target):
    """
    Check if the given prefix is a valid start of the target word or if the target word is a valid start of the given prefix.

    Args:
    prefix (str): The string prefix to be checked.
    target (str): The target word to compare the prefix against.

    Returns:
    bool: True if prefix is a valid start of target or if target is a valid start of prefix, False otherwise.
    """

    # Check if the target word starts with the given prefix.
    # This covers the case where the prefix is shorter than the target word.
    if target.startswith(prefix):
        return True

    # Check if the given prefix starts with the target word.
    # This covers the case where the prefix is longer than or equal to the target word.
    if prefix.startswith(target):
        return True

    # If neither of the above conditions are true, return False.
    return False


if __name__ == "__main__":
    import logging

    # Configure logging
    logging.basicConfig(level=logging.INFO)
    from transformers import AutoTokenizer

    tokenizer = AutoTokenizer.from_pretrained("gpt2", fast=True)

    trie = ByteTrie.from_tokenizer(tokenizer)
    print(f"length of trie: {len(trie)}=={len(tokenizer.vocab.items())}")

    trie.visualize(max_depth=0)

    #
    # print(trie.search("hello"))  # Example, replace with actual words from the vocab
    # print(trie.start_with_prefix("hell"))
    #
    # # Example Usage
    # words = trie.dfs(accept=lambda x: len(x) > 0 and x[0] == 65 or len(x)==0)
    # for word in words:
    #     print(bytes(word[0]).decode("utf-8"))
    #
    # # Example Usage
    # words = trie.bfs(predicate=lambda x: len(x) > 0 and x[0] == 65 or len(x)==0)
    # for word in words:
    #     print(bytes(word[0]).decode("utf-8"))
    #
    # token_acceptance = trie.get_next_token_acceptance(accept=lambda x: len(x) > 0 and x[0] == 65 or len(x)==0)
    # print(sum(token_acceptance))
    # assert sum(token_acceptance) == len(words)

    ########################
    # UTF-8
    ########################

    # from transformers import AutoTokenizer
    #
    # japanese = "こんにちは世界"
    # with open("examples/grammars/japanese.ebnf", "r") as file:
    #     input_text = file.read()
    # parsed_grammar = parse_ebnf(input_text)
    #
    # start_rule_id = parsed_grammar.symbol_table["root"]
    #
    # recognizer = GrammarRecognizer(parsed_grammar.grammar_encoding, start_rule_id)
    # parsing_state = recognizer.init_parsing_state()
    # token_acc = trie.get_next_token_acceptance(accept=lambda x: recognizer._probe_bytes_partial_match(x, parsing_state=parsing_state))



import unittest
import subprocess
import os
import glob


class TestExampleScripts(unittest.TestCase):
    def run_script(self, path_to_script):
        """Utility method to run a script using subprocess and assert that it exits with a 0 status code."""
        result = subprocess.run(
            ["python", path_to_script], capture_output=True, text=True
        )
        self.assertEqual(
            result.returncode,
            0,
            f"Script {path_to_script} failed with output:\n{result.stdout}\n{result.stderr}",
        )

    @unittest.skipIf(
        not os.environ.get("GITHUB_ACTIONS", False),
        "Skipping example tests because they are slow for local testing",
    )
    def test_all_example_scripts(self):
        """Test that all example scripts can be run without error."""
        for script_path in glob.glob("examples/*.py"):
            self.run_script(script_path)



from unittest import TestCase
from transformers import AutoTokenizer
from transformers_cfg.tokenization.utils import get_tokenizer_charset


class TestGetTokenizerCharset(TestCase):
    def test_get_tokenizer_charset(self):
        tokenizer = AutoTokenizer.from_pretrained("bert-base-uncased")
        charset = get_tokenizer_charset(tokenizer)
        expected_charset = set(
            "abcdefghijklmnopqrstuvwxyz0123456789!#$%&'()*+,-./:;<=>?@[\]^_`{|}~"
        )
        self.assertEqual(charset, expected_charset)



import unittest

from transformers import AutoTokenizer

from tests.test_accept_token_sequence._test_accept_tokens_mixin import (
    TokenizerTesterMixin,
)


# @unittest.skip("CodeGen is not supported and will be removed")
class DeepSeekTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = AutoTokenizer
    pretrained_name = "deepseek-ai/deepseek-coder-1.3b-base"

    def setUp(self):
        super().setUp()



from .token_grammar_recognizer import (
    IncrementalTokenRecognizer,
    NonIncrementalTokenSeqRecognizer,
)


# Old class name, kept for backward compatibility
IncrementalGrammarConstraint = IncrementalTokenRecognizer

NonIncrementalGrammarConstraint = NonIncrementalTokenSeqRecognizer



import unittest

from transformers import CodeGenTokenizerFast

from tests.test_accept_token_sequence._test_accept_tokens_mixin import (
    TokenizerTesterMixin,
)


# @unittest.skip("CodeGen is not supported and will be removed")
class PhiTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = CodeGenTokenizerFast
    pretrained_name = "microsoft/phi-1_5"

    def setUp(self):
        super().setUp()



import inspect
from typing import Dict
from transformers_cfg.tokenization.SUPPORTED_TOKENIZERS import SUPPORTED_TOKENIZERS


def replace_hex(match):
    hex_value = match.group(1)
    return chr(int(hex_value, 16))


# This will collect all imported classes from the current module (globals())
def get_imported_tokenizer_classes(module_globals) -> Dict[str, type]:
    return {
        name: obj
        for name, obj in module_globals.items()
        if inspect.isclass(obj) and name.endswith("TokenizerFast")
    }


def get_tokenizer_real_class(hf_tokenizer):
    return hf_tokenizer.__class__


def is_tokenizer_supported(hf_tokenizer_or_name):
    if isinstance(hf_tokenizer_or_name, str):
        from transformers import AutoTokenizer

        hf_tokenizer = AutoTokenizer.from_pretrained(hf_tokenizer_or_name)
    else:
        hf_tokenizer = hf_tokenizer_or_name
    return get_tokenizer_real_class(hf_tokenizer) in SUPPORTED_TOKENIZERS


def get_tokenizer_charset(hf_tokenizer):
    return set("".join(hf_tokenizer.vocab.keys()))



from transformers import (
    GPT2TokenizerFast,
    BartTokenizerFast,
    LlamaTokenizerFast,
    T5TokenizerFast,
    CodeGenTokenizerFast,
    PreTrainedTokenizerFast,
    GemmaTokenizerFast
)

SUPPORTED_TOKENIZERS = {
    GPT2TokenizerFast,
    BartTokenizerFast,
    LlamaTokenizerFast,
    T5TokenizerFast,
    CodeGenTokenizerFast,
    PreTrainedTokenizerFast,
    GemmaTokenizerFast
}



from unittest import TestCase


from tests.json_utils import is_json_parsable

from transformers_cfg.parser import parse_ebnf
from transformers_cfg.recognizer import StringRecognizer


class TestJsonArray(TestCase):
    def test_minimal_json_array(self):
        """
        Test that we can load a JSON array
        """
        # json = '["foo", {"bar":["baz", null, 1.0, 2]}]'
        jsons = [
            "[\\n]",
            "[\\n1]",
            "[\\n1,2]",
            "[\\n1,2,3]",
        ]
        with open("examples/grammars/json_arr.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)

        start_rule_id = parsed_grammar.symbol_table["root"]

        recognizer = StringRecognizer(parsed_grammar.grammar_encoding, start_rule_id)

        for json in jsons:
            self.assertEqual(
                is_json_parsable(json),
                recognizer._accept_prefix(json),
                f"Failed on {json}",
            )



from unittest import TestCase

from transformers_cfg.recognizer import StringRecognizer

from transformers_cfg.parser import parse_ebnf

import logging


class TestUnicode(TestCase):
    def test_accept_japanese(self):
        """
        Test that we can accept japanese characters
        """

        japanese = "こんにちは世界"
        with open("examples/grammars/japanese.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)

        start_rule_id = parsed_grammar.symbol_table["root"]

        recognizer = StringRecognizer(parsed_grammar.grammar_encoding, start_rule_id)

        bytes_japanese = bytes(japanese, "utf-8")
        logging.debug(
            f"bytes_japanese: {bytes_japanese} of length {len(bytes_japanese)}"
        )
        # こんにちは世界

        head_bytes = bytes_japanese[:8]
        # partial_utf8 = PartialUTF8()
        parsing_state = recognizer._update_state_with_bytes(head_bytes)

        # non empty stack means that the bytes were accepted
        self.assertTrue(len(parsing_state.stacks) > 0)

    def test_accept_japanese_progressive(self):
        #######################
        # Now consider the case of progressive matching
        #######################

        japanese = "こんにちは世界"
        with open("examples/grammars/japanese.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)

        start_rule_id = parsed_grammar.symbol_table["root"]

        recognizer = StringRecognizer(parsed_grammar.grammar_encoding, start_rule_id)

        bytes_japanese = bytes(japanese, "utf-8")
        logging.debug(
            f"bytes_japanese: {bytes_japanese} of length {len(bytes_japanese)}"
        )

        byte_tokens = [bytes_japanese[i] for i in range(len(bytes_japanese))]
        # cast into bytes
        byte_tokens = [bytes([byte]) for byte in byte_tokens]

        parsing_state = recognizer.get_initial_parsing_state()

        # parsing_state = recognizer.init_parsing_state
        for i, byte in enumerate(byte_tokens):
            parsing_state = recognizer._update_state_with_bytes(byte, parsing_state)
            self.assertTrue(len(parsing_state.stacks) > 0)

    def test_accept_emoji(self):
        """
        Test that we can accept emoji
        """

        emoji = "😀😄😂"
        with open("examples/grammars/emoji.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)

        start_rule_id = parsed_grammar.symbol_table["root"]

        recognizer = StringRecognizer(parsed_grammar.grammar_encoding, start_rule_id)

        bytes_emoji = bytes(emoji, "utf-8")
        logging.debug(f"bytes_emoji: {bytes_emoji} of length {len(bytes_emoji)}")
        # 😀😄😂

        # partial_utf8 = PartialUTF8()
        parsing_state = recognizer._update_state_with_bytes(bytes_emoji)
        # non empty stack means that the bytes were accepted
        self.assertTrue(len(parsing_state.stacks) > 0)



from setuptools import setup, find_packages

setup(
    name="transformers_cfg",
    version="v0.2.2",
    author="EPFL-dlab",
    author_email="saibo.geng@epfl.ch",
    description="Extension of Transformers library for Context-Free Grammar Constrained Decoding with EBNF grammars",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/epfl-dlab/transformers-CFG",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=open("requirements.txt").read().splitlines(),
    package_data={
        "transformers_cfg": ["examples/grammars/*.ebnf"],
    },
    include_package_data=True,
    # Add any additional package configuration here
    entry_points={
        "console_scripts": [
            "transformers-cfg-cli=transformers_cfg.cli.cli_main:main",
        ],
    },
)



from unittest import TestCase

from transformers_cfg.recognizer import StringRecognizer

from transformers_cfg.parser import parse_ebnf


class TestUnicode(TestCase):
    def test_accept_japanese(self):
        """
        Test that we can accept japanese characters
        """

        japanese = "こんにちは世界"
        with open("examples/grammars/japanese.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)

        start_rule_id = parsed_grammar.symbol_table["root"]

        recognizer = StringRecognizer(parsed_grammar.grammar_encoding, start_rule_id)

        self.assertTrue(recognizer._accept_prefix(japanese))

    def test_emoji(self):
        """
        Test that we can accept emoji
        """

        emoji = "😀😄😂"
        with open("examples/grammars/emoji.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)

        start_rule_id = parsed_grammar.symbol_table["root"]

        recognizer = StringRecognizer(parsed_grammar.grammar_encoding, start_rule_id)

        self.assertTrue(recognizer._accept_prefix(emoji))



import unittest

from transformers import BartTokenizerFast

from tests.test_accept_token_sequence._test_accept_tokens_mixin import (
    TokenizerTesterMixin,
)


class BartTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = BartTokenizerFast
    pretrained_name = "facebook/bart-large"

    def setUp(self):
        super().setUp()

    @unittest.skip(
        "BartTokenizer fails on emoji for some reason, need to investigate if user really needs this"
    )
    def test_emoji(self):
        super().test_emoji()



import re
from typing import List, Set
from transformers import (
    GPT2TokenizerFast,
    BartTokenizerFast,
    LlamaTokenizerFast,
    T5TokenizerFast,
    CodeGenTokenizerFast,
    PreTrainedTokenizerFast,
    GemmaTokenizerFast
)

from transformers_cfg.tokenization.SUPPORTED_TOKENIZERS import SUPPORTED_TOKENIZERS
from transformers_cfg.tokenization.utils import (
    replace_hex,
)


def get_TCFG_tokenizer_class(model_name_or_tokenizer):
    from transformers import AutoTokenizer

    if isinstance(model_name_or_tokenizer, str):
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_tokenizer)
    else:
        tokenizer = model_name_or_tokenizer

    return TCFG_Tokenizer.from_hf_tokenizer(tokenizer).__class__


class TCFG_Tokenizer:
    def __init__(self, hf_tokenizer):
        self.hf_tokenizer = hf_tokenizer
        self.special_token_ids = set(hf_tokenizer.all_special_ids)

    def real_vocab_size(self):
        return len(self.hf_tokenizer.vocab)

    def get_tokens_as_bytes(self) -> List[bytes]:
        vocab_size = self.real_vocab_size()
        token_as_bytes: List[bytes] = [
            self._format_token_as_bytes(i) for i in range(vocab_size)
        ]

        return token_as_bytes

    @classmethod
    def from_hf_tokenizer(cls, hf_tokenizer):
        assert (
            type(hf_tokenizer) in SUPPORTED_TOKENIZERS
        ), f"Tokenizer not supported: {hf_tokenizer.__class__.__name__}, supported tokenizers: {SUPPORTED_TOKENIZERS}"

        if isinstance(
            hf_tokenizer,
            (GPT2TokenizerFast, BartTokenizerFast),
        ):
            return TCFG_GPT2Tokenizer(hf_tokenizer)
        elif isinstance(hf_tokenizer, (LlamaTokenizerFast, GemmaTokenizerFast, T5TokenizerFast)):
            return TCFG_LlamaTokenizer(hf_tokenizer)
        elif isinstance(hf_tokenizer, CodeGenTokenizerFast):
            # phi reuses the codegen tokenizer
            return TCFG_PhiTokenizer(hf_tokenizer)
        elif isinstance(
            hf_tokenizer, PreTrainedTokenizerFast
        ) and hf_tokenizer.name_or_path.startswith("meta-llama/Meta-Llama-3"):
            return TCFG_LlamaTokenizer(hf_tokenizer)
        else:
            raise NotImplementedError(
                f"Tokenizer not supported: {hf_tokenizer.__class__.__name__}"
            )

    # will be extended by the subclasses
    def get_special_token_ids_to_excluded(self) -> Set[int]:
        return self.special_token_ids


class TCFG_LlamaTokenizer(TCFG_Tokenizer):
    def __init__(self, hf_tokenizer):
        super().__init__(hf_tokenizer)

    def _format_token_as_bytes(self, token_id):
        token = self.hf_tokenizer.convert_ids_to_tokens(token_id)
        token = re.sub(r"<0x([0-9a-fA-F]{2})>", replace_hex, token)
        # token = token.replace("▁", " ")
        return bytes(token, "utf-8")

    def get_special_token_ids_to_excluded(self):
        if self.hf_tokenizer.name_or_path.startswith("deepseek-ai/deepseek-coder"):
            # deepseek has in total 22 special tokens, with token_ids from 32000 to 32021
            # with first 13 being characters for bytes: {'õ': 32000, '÷': 32001, 'Á': 32002, 'ý': 32003, 'À': 32004, 'ÿ': 32005, 'ø': 32006, 'ú': 32007, 'þ': 32008, 'ü': 32009, 'ù': 32010, 'ö': 32011, 'û': 32012}
            # the rest are special tokens for the tokenizer: { '<｜begin▁of▁sentence｜>': 32013, '<｜end▁of▁sentence｜>': 32014, '<｜fim▁hole｜>': 32015, '<｜fim▁begin｜>': 32016, '<｜fim▁end｜>': 32017, '<pad>': 32018, '<|User|>': 32019, '<|Assistant|>': 32020, '<|EOT|>': 32021}
            added_vocab_dict = self.hf_tokenizer.get_added_vocab()
            added_tokens_id_to_excluded = set(
                [
                    token_id
                    for tok, token_id in added_vocab_dict.items()
                    if tok.startswith("<｜")
                ]
            )
            return self.special_token_ids.union(added_tokens_id_to_excluded)
        return self.special_token_ids


class TCFG_GPT2Tokenizer(TCFG_Tokenizer):
    def __init__(self, hf_tokenizer):
        super().__init__(hf_tokenizer)

    def _format_token_as_bytes(self, token_id):
        if token_id in self.special_token_ids:
            return None
        return bytes(
            self.hf_tokenizer.decode([token_id], clean_up_tokenization_spaces=False),
            "utf-8",
        )


class TCFG_CharacterTokenizer(TCFG_Tokenizer):
    """
    Not yet used, but can be used for character level tokenization (even though rarely used in practice)
    """

    def __init__(self, hf_tokenizer):
        super().__init__(hf_tokenizer)

    def _format_token_as_bytes(self, token_id):
        return bytes(self.hf_tokenizer.convert_ids_to_tokens(token_id), "utf-8")


class TCFG_PhiTokenizer(TCFG_GPT2Tokenizer):
    def __init__(self, hf_tokenizer):
        super().__init__(hf_tokenizer)

    def real_vocab_size(self):
        return 50257  # 50 k tokens + 256 for bytes + 1 for EOS
